import logging
from typing import List, Optional
from xml.etree import ElementTree as ET

from aws.osml.photogrammetry import (
    RSMContext,
    RSMGroundDomain,
    RSMGroundDomainForm,
    RSMImageDomain,
    RSMLowOrderPolynomial,
    RSMPolynomial,
    RSMPolynomialSensorModel,
    RSMSectionedPolynomialSensorModel,
    SensorModel,
    WorldCoordinate,
)

from .sensor_model_builder import SensorModelBuilder
from .xmltre_utils import get_tre_field_value, parse_rpc_coefficients


class RSMSensorModelBuilder(SensorModelBuilder):
    """
    This builder is used to create sensor models for images that have RSM TREs. The inputs are not the TREs themselves
    but rather the XML formatted TREs resulting from GDAL's parsing of that information. This information can be
    obtained by accessing the xml:TRE metadata domain for a given GDAL dataset.

    The actual type and number of RSM TREs included with an image will vary depending on the type of RSM sensor model
    defined. In general all images with these sensor models must have an RSMID TRE that defines the overall context
    of the sensor model along with RSMDCA and RSMECA TREs that define the direct and indirect error covariance
    data.

    The polynomial based sensor models will then have at least one RSMPCA TRE and may have multiple. If there are
    multiple then a RSMPIA TRE will also be present to describe how the various polynomial models cover the overall
    image domain.

    The grid based sensor models will have at least one RSMGGA TRE and may have multiple. If there are multiple then
    a RSMGIA TRE will also be present to describe how the various interpolation grids cover the overall image domain.

    It is also possible that adjustment parameters will be in RSMAPA TREs. These adjustments are frequently used to
    modify / update a sensor model after it was originally created.

    See STDI-0002 Volume 1 Appendix U for more detailed information.
    """

    def __init__(self, xml_tres: ET.Element) -> None:
        """
        Constructor for the builder accepting the required XML TREs.

        :param xml_tres: the XML tres for this image

        :return: None
        """
        super().__init__()
        self.xml_tres = xml_tres

    def build(self) -> Optional[SensorModel]:
        """
        Examine the TRE metadata for RSM information, parse the necessary values out of those TREs, and construct a
        RSM sensor model.

        :return: a RSM Polynomial SensorModel if one can be constructed, None otherwise
        """

        # Check to see if an RSMIDA TRE is included with the metadata. This is a mandatory TRE that will be available
        # for all RSM based sensor models. If it is not available then we can stop because the metadata does not
        # contain the basic information necessary to support RSM.
        rsmid_tre = self.xml_tres.find("./tre[@name='RSMIDA']")
        if rsmid_tre is None:
            logging.debug("No RSMID TRE found. Skipping RSM sensor model build.")
            return None

        try:
            # Use the information in the RSMID TRE to build the context. This context contains the ground and image
            # domains that bound the valid regions for the RSM model in this image.
            rsm_context = RSMSensorModelBuilder._build_rsm_context(rsmid_tre)

            # If an RSM model is using polynomial coefficients to define the sensor model then those coefficients
            # will be stored in RSMPC TREs. Note that some images may have several of these TREs defined so we will
            # construct RSMPolynomialSensorModels for every TRE of this kind.
            rsmpc_tres = self.xml_tres.findall("./tre[@name='RSMPCA']")
            rsm_polynomial_sensor_models = [
                RSMSensorModelBuilder._build_rsm_polynomial_sensor_model(rsmpc_tre, rsm_context) for rsmpc_tre in rsmpc_tres
            ]

            # If we only have one RSM polynomial sensor model then it applies to the entire RSM domain. If we have
            # multiple then we are dealing with a sectioned sensor model which will require additional TREs to be
            # parsed.
            if len(rsm_polynomial_sensor_models) > 0:
                if len(rsm_polynomial_sensor_models) == 1:
                    return rsm_polynomial_sensor_models[0]
                else:
                    # Parse RSMPI and construct a segmented polynomial sensor model
                    rsmpi_tre = self.xml_tres.find("./tre[@name='RSMPIA']")
                    if rsmpi_tre is None:
                        logging.warning(
                            "Image has multiple RSMPCA TREs but is missing a RSMPIA that assigns them to sections."
                            "No sensor model can be built! "
                        )
                        return None
                    return RSMSensorModelBuilder._build_rsm_sectioned_polynomial_sensor_model(
                        rsmpi_tre, rsm_context, rsm_polynomial_sensor_models
                    )

            # TODO: Check for RSMGGA and RSMGIA TREs and construct a RSM grid interpolation sensor model
            logging.warning(
                "Image has RSMID TRE but no polynomials. Grid based RSM not implemented so no sensor model returned."
            )
            return None

        except ValueError as ve:
            logging.warning("Unable to parse RSM TREs found in XML metadata. No SensorModel created.")
            logging.warning(str(ve))

            return None

    @staticmethod
    def _build_rsm_ground_domain(rsmid_tre: ET.Element) -> RSMGroundDomain:
        """
        This private method constructs the ground domain from information in the RSMIDA TRE.

        :param rsmid_tre: the GDAL XML for RSMIDA

        :return: the ground domain
        """

        # This is the type of ground domain for this RSM sensor model
        ground_domain_form = RSMGroundDomainForm(get_tre_field_value(rsmid_tre, "GRNDD", str))

        # The valid region of the ground domain is defined by 8 sets of world coordinates (V1 - V8)
        ground_domain_vertices = [
            WorldCoordinate(
                [
                    get_tre_field_value(rsmid_tre, f"V{vertex_number}X", float),
                    get_tre_field_value(rsmid_tre, f"V{vertex_number}Y", float),
                    get_tre_field_value(rsmid_tre, f"V{vertex_number}Z", float),
                ]
            )
            for vertex_number in range(1, 9)
        ]

        # Ground domains are either rectangular or geodetic. Rectangular ground domains have additional values
        # that define a cartesian coordinate system anchored at a point on the earth.
        rectangular_coordinate_origin = None
        rectangular_coordinate_unit_vectors = None
        if ground_domain_form == RSMGroundDomainForm.RECTANGULAR:
            # The world location for the origin (0, 0, 0) of the rectangular coordinate system
            rectangular_coordinate_origin = WorldCoordinate(
                [
                    get_tre_field_value(rsmid_tre, "XUOR", float),
                    get_tre_field_value(rsmid_tre, "YUOR", float),
                    get_tre_field_value(rsmid_tre, "ZUOR", float),
                ]
            )

            # Unit vectors defining the cartesian coordinate system
            rectangular_coordinate_unit_vectors = []
            for row_coefficient in ["XR", "YR", "ZR"]:
                row = [
                    get_tre_field_value(rsmid_tre, f"{col_coefficient}U{row_coefficient}", float)
                    for col_coefficient in ["X", "Y", "Z"]
                ]
                rectangular_coordinate_unit_vectors.append(row)

        try:
            ground_reference_point = WorldCoordinate(
                [
                    get_tre_field_value(rsmid_tre, "GRPX", float),
                    get_tre_field_value(rsmid_tre, "GRPY", float),
                    get_tre_field_value(rsmid_tre, "GRPZ", float),
                ]
            )
        except ValueError:
            # The ground reference point is optional and these elements may be filled with spaces. If we can't parse
            # floating point numbers from these fields we should assume this information has not been provided.
            ground_reference_point = None

        return RSMGroundDomain(
            ground_domain_form,
            ground_domain_vertices,
            rectangular_coordinate_origin=rectangular_coordinate_origin,
            rectangular_coordinate_unit_vectors=rectangular_coordinate_unit_vectors,
            ground_reference_point=ground_reference_point,
        )

    @staticmethod
    def _build_rsm_image_domain(rsmid_tre: ET.Element) -> RSMImageDomain:
        """
        This private method constructs the image domain from information in the RSMIDA TRE.

        :param rsmid_tre: the GDAL XML for RSMIDA
        :return: the image domain
        """
        return RSMImageDomain(
            get_tre_field_value(rsmid_tre, "MINR", int),
            get_tre_field_value(rsmid_tre, "MAXR", int),
            get_tre_field_value(rsmid_tre, "MINC", int),
            get_tre_field_value(rsmid_tre, "MAXC", int),
        )

    @staticmethod
    def _build_rsm_context(rsmid_tre: ET.Element) -> RSMContext:
        """
        This private method constructs a RSM context from information in the RSMIDA TRE.

        :param rsmid_tre: the GDAL XML for RSMIDA

        :return: the RSM context
        """
        return RSMContext(
            RSMSensorModelBuilder._build_rsm_ground_domain(rsmid_tre),
            RSMSensorModelBuilder._build_rsm_image_domain(rsmid_tre),
        )

    @staticmethod
    def _build_rsm_polynomial(rsmpc_tre: ET.Element, polynomial_prefix: str) -> RSMPolynomial:
        """
        This private method constructs an RSM polynomial from a group of related fields in the RSMPCA TRE. These
        TREs have similar fields grouped by the RN, RD, CN, and CD prefixes which correspond to the row or column
        (R or C) numerator or denominator (N or D) identifiers for the polynomial they're associated with.

        :param rsmpc_tre: the GDAL XML for RSMPCA
        :param polynomial_prefix: the prefix identifying the polynomial

        :return: RSMPolynomial = the RSM polynomial
        """
        if polynomial_prefix not in ["RN", "RD", "CN", "CD"]:
            raise ValueError(f"Unexpected prefix {polynomial_prefix}. Expecting RN, RD, CN, or CD")
        return RSMPolynomial(
            get_tre_field_value(rsmpc_tre, f"{polynomial_prefix}PWRX", int),
            get_tre_field_value(rsmpc_tre, f"{polynomial_prefix}PWRY", int),
            get_tre_field_value(rsmpc_tre, f"{polynomial_prefix}PWRZ", int),
            parse_rpc_coefficients(rsmpc_tre, f"{polynomial_prefix}PCF"),
        )

    @staticmethod
    def _build_rsm_polynomial_sensor_model(rsmpc_tre: ET.Element, rsm_context: RSMContext) -> RSMPolynomialSensorModel:
        """
        This private method constructs an RSM polynomial sensor model from an RSMPCA TRE and the context object.

        :param rsmpc_tre: the GDAL XML for RSMPCA
        :param rsm_context: the corresponding RSM context

        :return: the RSM polynomial sensor model
        """
        return RSMPolynomialSensorModel(
            rsm_context,
            get_tre_field_value(rsmpc_tre, "RSN", int),
            get_tre_field_value(rsmpc_tre, "CSN", int),
            get_tre_field_value(rsmpc_tre, "RNRMO", float),
            get_tre_field_value(rsmpc_tre, "CNRMO", float),
            get_tre_field_value(rsmpc_tre, "XNRMO", float),
            get_tre_field_value(rsmpc_tre, "YNRMO", float),
            get_tre_field_value(rsmpc_tre, "ZNRMO", float),
            get_tre_field_value(rsmpc_tre, "RNRMSF", float),
            get_tre_field_value(rsmpc_tre, "CNRMSF", float),
            get_tre_field_value(rsmpc_tre, "XNRMSF", float),
            get_tre_field_value(rsmpc_tre, "YNRMSF", float),
            get_tre_field_value(rsmpc_tre, "ZNRMSF", float),
            RSMSensorModelBuilder._build_rsm_polynomial(rsmpc_tre, "RN"),
            RSMSensorModelBuilder._build_rsm_polynomial(rsmpc_tre, "RD"),
            RSMSensorModelBuilder._build_rsm_polynomial(rsmpc_tre, "CN"),
            RSMSensorModelBuilder._build_rsm_polynomial(rsmpc_tre, "CD"),
        )

    @staticmethod
    def _build_loworder_rsm_polynomial(rsmpi_tre: ET.Element, polynomial_prefix: str) -> RSMLowOrderPolynomial:
        """
        This private method constructs a low order RSM polynomial from a group of related fields in the RSMPIA TRE.
        These TREs have similar fields grouped by the R and C prefixes which correspond to the row or column
        identifiers for the polynomial they're associated with.

        :param rsmpi_tre: the GDAL XML for RSMPIA
        :param polynomial_prefix: the prefix identifying the polynomial

        :return: the low order RSM polynomial
        """
        if polynomial_prefix not in ["R", "C"]:
            raise ValueError(f"Unexpected prefix {polynomial_prefix}. Expecting R or C")

        coefficients = []
        for coeff_suffix in ["0", "X", "Y", "Z", "XX", "XY", "XZ", "YY", "YZ", "ZZ"]:
            coefficients.append(get_tre_field_value(rsmpi_tre, f"{polynomial_prefix}{coeff_suffix}", int))
        return RSMLowOrderPolynomial(coefficients)

    @staticmethod
    def _build_rsm_sectioned_polynomial_sensor_model(
        rsmpi_tre: ET.Element,
        rsm_context: RSMContext,
        rsm_polynomial_sensor_models: List[RSMPolynomialSensorModel],
    ) -> RSMSectionedPolynomialSensorModel:
        """
        This private method constructs an RSM sectioned polynomial sensor model from an RSMPIA TRE, the context object,
        and a collection of RSMPolynomialSensorModels.

        :param rsmpi_tre: the GDAL XML for RSMPCA
        :param rsm_context: the corresponding RSM context


        :return: the RSM polynomial sensor model
        """
        num_section_rows = get_tre_field_value(rsmpi_tre, "RNIS", int)
        num_section_cols = get_tre_field_value(rsmpi_tre, "CNIS", int)
        sensor_model_grid_map = {}
        for sensor_model in rsm_polynomial_sensor_models:
            sensor_model_grid_map[(sensor_model.section_row, sensor_model.section_col)] = sensor_model

        section_sensor_model_grid = []
        for row in range(0, num_section_rows):
            row_of_sensor_models = []
            for col in range(0, num_section_cols):
                row_of_sensor_models.append(sensor_model_grid_map[(row, col)])
            section_sensor_model_grid.append(row_of_sensor_models)

        return RSMSectionedPolynomialSensorModel(
            rsm_context,
            num_section_rows,
            num_section_cols,
            get_tre_field_value(rsmpi_tre, "RSSIZ", float),
            get_tre_field_value(rsmpi_tre, "CSSIZ", float),
            RSMSensorModelBuilder._build_loworder_rsm_polynomial(rsmpi_tre, "R"),
            RSMSensorModelBuilder._build_loworder_rsm_polynomial(rsmpi_tre, "C"),
            section_sensor_model_grid,
        )



from math import sqrt
from typing import List


def equilateral_triangle(centroid: List[float], size: float) -> List[List[float]]:
    """
    Utility function to create an equilateral triangle with a given side length centered on a point. There
    are many solutions to that problem so this returns the solution where the top corner is directly above
    the center and the base are horizontal. (i.e. for a triangle made of points ABC A[0] == center[0] and
    B[1] == C[1])

    :param centroid: the center of the triangle
    :param size: the length of one side

    :return: a list of lists representing the triangle len(result) == 3, len(result[0]) == 2
    """
    sized_triangle_at_origin = [
        [0.0, size * sqrt(3.0) / 3.0],
        [-1.0 * size / 2.0, -1.0 * sqrt(3) / 6.0 * size],
        [size / 2.0, -1.0 * sqrt(3) / 6.0 * size],
    ]

    return [[centroid[0] + coord[0], centroid[1] + coord[1]] for coord in sized_triangle_at_origin]



import logging
import re
from typing import Callable, List, TypeVar
from xml.etree import ElementTree as ET

from defusedxml import ElementTree

logger = logging.getLogger(__name__)

# This is a type placeholder needed by the parse_element_text() type hints
T = TypeVar("T")


class SICDUpdater:
    """
    This class provides a means to perform common updates to a SICD XML metadata document.
    """

    def __init__(self, sicd_element: ET.Element):
        """
        Construct a new instance of this class to manage a given set of SICD metadata.

        :param sicd_element: the SICD XML metadata to update
        """
        self.sicd_element = sicd_element

        # Extract the XML namespace from the root SICD element and store it for later use in element queries
        namespace_match = re.match(r"{.*}", self.sicd_element.tag)
        self.namespace = namespace_match.group(0) if namespace_match else ""

        # We don't currently have many examples of SICD data. An attempt has been made to make this code
        # work so long as the portions of the XML schema we depend upon don't change. This warning is just
        # an attempt to provide diagnostic information incase future datasets don't work.
        if self.namespace != "{urn:SICD:1.2.1}":
            logger.warning(f"Attempting to process SICD metadata with an untested namespace {self.namespace}")

        # Here we're storing off the original first row/col to support the case where multiple chips are
        # created from a SICD image that has already been chipped.
        self.original_first_row = self.parse_element_text(".//{0}FirstRow".format(self.namespace), int)
        self.original_first_col = self.parse_element_text(".//{0}FirstCol".format(self.namespace), int)

    def update_image_data_for_chip(self, chip_bounds: List[int]) -> None:
        """
        This updates the SICD ImageData structure so that the FirstRow, FirstCol and NumRows, NumCols
        elements match the new chip boundary.A sample of this XML structure is shown below::

            <ImageData>
                <NumRows>3859</NumRows>
                <NumCols>6679</NumCols>
                <FirstRow>0</FirstRow>
                <FirstCol>0</FirstCol>
                <FullImage>
                    <NumRows>3859</NumRows>
                    <NumCols>6679</NumCols>
                </FullImage>
            </ImageData>


        :param chip_bounds: the [col, row, width, height] of the chip boundary
        """
        first_col_element = self.sicd_element.find(".//{0}FirstCol".format(self.namespace))
        first_row_element = self.sicd_element.find(".//{0}FirstRow".format(self.namespace))
        num_cols_element = self.sicd_element.find(".//{0}NumCols".format(self.namespace))
        num_rows_element = self.sicd_element.find(".//{0}NumRows".format(self.namespace))
        if first_row_element is None or first_col_element is None or num_cols_element is None or num_rows_element is None:
            logger.warning("SICD ImageData structures were not found. No updates applied.")
            return

        first_col_element.text = str(self.original_first_col + chip_bounds[0])
        first_row_element.text = str(self.original_first_row + chip_bounds[1])
        num_cols_element.text = str(chip_bounds[2])
        num_rows_element.text = str(chip_bounds[3])

        if logger.isEnabledFor(logging.DEBUG):
            image_data_element = self.sicd_element.find(".//{0}ImageData".format(self.namespace))
            if image_data_element is not None:
                logger.debug("Updated SICD ImageData element for chip:")
                logger.debug(
                    ElementTree.tostring(
                        image_data_element,
                        encoding="unicode",
                    )
                )

    def encode_current_xml(self) -> str:
        """
        Returns a copy of the current SICD metadata encoded in XML.

        :return: xml encoded SICD metadata
        """
        return ElementTree.tostring(self.sicd_element, encoding="unicode")

    def parse_element_text(self, element_xpath: str, type_conversion: Callable[[str], T]) -> T:
        """
        This function finds the first element matching the provided xPath and then runs the text of that element
        through the provided conversion function.

        :param element_xpath: the xPath of the element
        :param type_conversion: the desired type of the output, must support construction from a string
        :return: the element text converted to the requested type
        """
        field_element = self.sicd_element.find(element_xpath)
        if field_element is None:
            raise ValueError(f"Unable to find element {element_xpath}")
        str_value = field_element.text
        if str_value is None:
            raise ValueError(f"Element {element_xpath} does not have text.")
        return type_conversion(str_value)



# Telling flake8 to not flag errors in this file. It is normal that these classes are imported but not used in an
# __init__.py file.
# flake8: noqa
"""
The image_processing package contains various utilities for manipulating overhead imagery.
"""

from .gdal_tile_factory import GDALTileFactory

__all__ = ["GDALTileFactory"]



# TODO: Add typing for ArrayLike once Numpy upgraded to 1.20+
# from numpy.typing import ArrayLike

import operator
from abc import ABC, abstractmethod
from typing import Any, Optional, Tuple

from cachetools import LRUCache, cachedmethod
from scipy.interpolate import RectBivariateSpline

from .coordinates import GeodeticWorldCoordinate
from .elevation_model import ElevationModel, ElevationRegionSummary
from .sensor_model import SensorModel


class DigitalElevationModelTileSet(ABC):
    """
    This class defines an abstraction that is capable of identifying which elevation tile in a
    DEM contains the elevations for a given world coordinate. It is common to split a DEM with
    global coverage up into a collection of files with well understood coverage areas. Those files
    may follow a simple naming convention but could be cataloged in an external spatial index. This
    class abstracts those details away from the DigitalElevationModel allowing us to easily extend
    this design to various tile sets.

    :return: None
    """

    def __init__(self) -> None:
        pass

    @abstractmethod
    def find_tile_id(self, geodetic_world_coordinate: GeodeticWorldCoordinate) -> Optional[str]:
        """
        Converts the latitude, longitude of the world coordinate into a tile path.

        :param geodetic_world_coordinate: GeodeticWorldCoordinate = the world coordinate of interest

        :return: the tile path or None if the DEM does not have coverage for this location
        """


class DigitalElevationModelTileFactory(ABC):
    """
    This class defines an abstraction that is able to load a tile and convert it to a Numpy array
    of elevation data along with a SensorModel that can be used to identify the grid locations
    associated with a latitude, longitude.

    :return: None
    """

    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_tile(self, tile_path: str) -> Tuple[Optional[Any], Optional[SensorModel], Optional[ElevationRegionSummary]]:
        """
        Retrieve a numpy array of elevation values and a sensor model.

        TODO: Replace Any with numpy.typing.ArrayLike once we move to numpy >1.20

        :param tile_path: the location of the tile to load

        :return: an array of elevation values, a sensor model, and a summary
        """


class DigitalElevationModel(ElevationModel):
    """
    A Digital Elevation Model (DEM) is a representation of the topographic surface of the Earth. Theoretically
    these representations exclude trees, buildings, and any other surface objects but in practice elevations
    from those objects are likely captured by the sensors use to capture the elevation data.

    These datasets are normally stored as a pre-tiled collection of images with a well established resolution and
    coverage.
    """

    def __init__(
        self,
        tile_set: DigitalElevationModelTileSet,
        tile_factory: DigitalElevationModelTileFactory,
        raster_cache_size: int = 10,
    ) -> None:
        """
        This constructor accepts a tile set and a tile factory which specify how to index into and
        load elevations from a TBD collection of elevation tiles that make up a DEM. This inversion of
        control pattern isolates the generic processing of interpolating elevation values from a regular
        grid away from the details of how the elevation tiles are stored and accessed.

        :param tile_set: DigitalElevationModelTileSet = a class used to identify the DEM tile associated with a location
        :param tile_factory: DigitalElevationModelTileFactory = a class used to load DEM tile and convert to numpy array
        :param raster_cache_size: int = the number of DEM arrays to store in memory preventing frequent loading

        :return: None
        """
        super().__init__()
        self.tile_set = tile_set
        self.tile_factory = tile_factory
        self.raster_cache: LRUCache = LRUCache(maxsize=raster_cache_size)
        # TODO: Think about raster_cache_size parameter. This is the number of rasters we will keep open at any
        #       one time. Look at the size of those tiles and add a comment about how much memory will be used by this
        #       setting. Pick a default that is reasonable and also likely to cover most images

    def set_elevation(self, geodetic_world_coordinate: GeodeticWorldCoordinate) -> None:
        """
        This method updates the elevation component of a geodetic world coordinate to match the surface
        elevation at the provided latitude and longitude. Note that if the DEM does not have elevation
        values for this region or if there is an error loading the associated image the elevation will
        be unchanged.

        :param geodetic_world_coordinate: the coordinate to update

        :return: None
        """
        tile_id = self.tile_set.find_tile_id(geodetic_world_coordinate)
        if not tile_id:
            return

        interpolation_grid, sensor_model, summary = self.get_interpolation_grid(tile_id)

        if interpolation_grid is not None and sensor_model is not None:
            image_coordinate = sensor_model.world_to_image(geodetic_world_coordinate)
            geodetic_world_coordinate.elevation = interpolation_grid(image_coordinate.x, image_coordinate.y)[0][0]

    def describe_region(self, geodetic_world_coordinate: GeodeticWorldCoordinate) -> Optional[ElevationRegionSummary]:
        """
        Get a summary of the region near the provided world coordinate

        :param geodetic_world_coordinate: the coordinate at the center of the region of interest
        :return: a summary of the elevation data in this tile
        """

        tile_id = self.tile_set.find_tile_id(geodetic_world_coordinate)
        if not tile_id:
            return

        interpolation_grid, sensor_model, summary = self.get_interpolation_grid(tile_id)
        return summary

    @cachedmethod(operator.attrgetter("raster_cache"))
    def get_interpolation_grid(
        self, tile_path: str
    ) -> Tuple[Optional[RectBivariateSpline], Optional[SensorModel], Optional[ElevationRegionSummary]]:
        """
        This method loads and converts an array of elevation values into a class that can
        interpolate values that lie between measured elevations. The sensor model is also
        returned to allow us to precisely identify the location in the grid of a
        world coordinate.

        Note that the results of this method are cached by tile_id. It is very common for
        the set_elevation() method to be called multiple times for locations that are in a
        narrow region of interest. This will prevent unnecessary repeated loading of tiles.

        :param tile_path: the location of the tile to load

        :return: the cached interpolation object, sensor model, and summary
        """
        elevations_array, sensor_model, summary = self.tile_factory.get_tile(tile_path)
        if elevations_array is not None and sensor_model is not None:
            height, width = elevations_array.shape
            x = range(0, width)
            y = range(0, height)
            return RectBivariateSpline(x, y, elevations_array.T, kx=1, ky=1), sensor_model, summary
        else:
            return None, None, None



import base64
import logging
from enum import Enum
from typing import List, Optional
from xml.etree import ElementTree as ET

from osgeo import gdal

from aws.osml.photogrammetry import ChippedImageSensorModel, CompositeSensorModel, ImageCoordinate, SensorModel

from .gdal_sensor_model_builder import GDALAffineSensorModelBuilder, GDALGCPSensorModelBuilder
from .nitf_des_accessor import NITFDESAccessor
from .projective_sensor_model_builder import ProjectiveSensorModelBuilder
from .rpc_sensor_model_builder import RPCSensorModelBuilder
from .rsm_sensor_model_builder import RSMSensorModelBuilder
from .sicd_sensor_model_builder import SICDSensorModelBuilder
from .xmltre_utils import get_tre_field_value


class ChippedImageInfoFacade:
    """
    This is a facade class that can be initialized with an ICHIPB TRE. It provides accessors for the values
    so that they can easily be used to create an ChippedImageSensorModel
    """

    def __init__(self, ichipb_tre: ET.Element) -> None:
        """
        Constructor initializes the properties from values in the TRE.

        :param ichipb_tre: the GDAL XML for the ICHIPB TRE

        :return: None
        """
        try:
            # Loop through the Output Product (OP) and Full Image (FI) fields in the ICHIPB TRE and construct
            # the corresponding image coordinates needed to create a chipped sensor model.
            self.full_image_coordinates = []
            self.chipped_image_coordinates = []
            for grid_point in ["11", "12", "21", "22"]:
                op_col = get_tre_field_value(ichipb_tre, f"OP_COL_{grid_point}", float)
                op_row = get_tre_field_value(ichipb_tre, f"OP_ROW_{grid_point}", float)
                fi_col = get_tre_field_value(ichipb_tre, f"FI_COL_{grid_point}", float)
                fi_row = get_tre_field_value(ichipb_tre, f"FI_ROW_{grid_point}", float)
                self.full_image_coordinates.append(ImageCoordinate([fi_col, fi_row]))
                self.chipped_image_coordinates.append(ImageCoordinate([op_col, op_row]))

            self.full_image_width = get_tre_field_value(ichipb_tre, "FI_COL", int)
            self.full_image_height = get_tre_field_value(ichipb_tre, "FI_ROW", int)
        except ValueError as ve:
            logging.warning("Unable to parse ICHIPB TRE found in XML metadata. SensorModel is unchanged.")
            logging.warning(str(ve))


class SensorModelTypes(Enum):
    """
    This enumeration defines the various sensor model types this factory can build.
    """

    AFFINE = "AFFINE"
    PROJECTIVE = "PROJECTIVE"
    RPC = "RPC"
    RSM = "RSM"
    SICD = "SICD"


ALL_SENSOR_MODEL_TYPES = [item for item in SensorModelTypes]


class SensorModelFactory:
    """
    This class encapsulates the logic necessary to construct SensorModels from imagery metadata parsed using GDAL.
    Users initialize the builder by providing whatever metadata is available and this class will decide how to create
    the most accurate SensorModel from the available information.
    """

    def __init__(
        self,
        actual_image_width: int,
        actual_image_height: int,
        xml_tres: Optional[ET.Element] = None,
        xml_dess: Optional[List[str]] = None,
        geo_transform: Optional[List[float]] = None,
        proj_wkt: Optional[str] = None,
        ground_control_points: Optional[List[gdal.GCP]] = None,
        selected_sensor_model_types: Optional[List[SensorModelTypes]] = None,
    ) -> None:
        """
        Construct a builder providing whatever metadata is available from the image. All of the parameters are named and
        optional allowing users to provide whatever they can and trusting that this builder will make use of as much of
        the information as possible.

        :param actual_image_width: width of the current image
        :param actual_image_height: height of the current image
        :param xml_tres: XML representing metadata in the tagged record extensions(TRE)
        :param xml_dess: XML representing data contained in the data extension segments (DES)
        :param geo_transform: a GDAL affine transform
        :param proj_wkt: the well known text string of the CRS used by the image
        :param ground_control_points: a list of GDAL GCPs that identify correspondences in the image
        :param selected_sensor_model_types: a list of sensor models that should be attempted by this factory

        :return: None
        """
        if selected_sensor_model_types is None:
            selected_sensor_model_types = ALL_SENSOR_MODEL_TYPES
        self.actual_image_width = actual_image_width
        self.actual_image_height = actual_image_height
        self.xml_tres = xml_tres
        self.xml_dess = xml_dess
        self.geo_transform = geo_transform
        self.proj_wkt = proj_wkt
        self.ground_control_points = ground_control_points
        self.selected_sensor_model_types = selected_sensor_model_types

    def build(self) -> Optional[SensorModel]:
        """
        Constructs the sensor model from the available information. Note that in cases where not enough information is
        available to provide any solution this method will return None.

        :return: the highest quality sensor model available given the information provided
        """

        approximate_sensor_model = None
        precision_sensor_model = None

        if SensorModelTypes.AFFINE in self.selected_sensor_model_types:
            if self.geo_transform is not None:
                approximate_sensor_model = GDALAffineSensorModelBuilder(self.geo_transform, self.proj_wkt).build()

        if SensorModelTypes.PROJECTIVE in self.selected_sensor_model_types:
            if self.ground_control_points is not None and len(self.ground_control_points) > 3:
                approximate_sensor_model = GDALGCPSensorModelBuilder(self.ground_control_points).build()

        if self.xml_tres is not None:
            # Start with the assumption that the raster we have is the full image. We will update this later if
            # it turns out we're working with an image chip.
            full_image_width = self.actual_image_width
            full_image_height = self.actual_image_height

            # Check to see if this image is a chip from a larger image and if so extract the chip corner information
            # from the ICHIPB TRE.
            chipped_image_info = None
            ichipb_tre = self.xml_tres.find("./tre[@name='ICHIPB']")
            if ichipb_tre is not None:
                chipped_image_info = ChippedImageInfoFacade(ichipb_tre)
                full_image_width = chipped_image_info.full_image_width
                full_image_height = chipped_image_info.full_image_height

            # Attempt to build a robust sensor model from either RSM or RPC metadata in the TREs. These
            # sensor models always reference the full image so if this is a chip we wrap the resulting sensor
            # model using information taken from ICHIPB. Note that in the unlikely event that an image has both
            # RSM and RPC metadata the RSM will be used because it has been developed as a replacement for RPC.
            precision_sensor_model = None
            if SensorModelTypes.RSM in self.selected_sensor_model_types:
                precision_sensor_model = RSMSensorModelBuilder(self.xml_tres).build()
            if precision_sensor_model is None and SensorModelTypes.RPC in self.selected_sensor_model_types:
                precision_sensor_model = RPCSensorModelBuilder(self.xml_tres).build()
            if precision_sensor_model is not None and chipped_image_info is not None:
                precision_sensor_model = ChippedImageSensorModel(
                    chipped_image_info.full_image_coordinates,
                    chipped_image_info.chipped_image_coordinates,
                    precision_sensor_model,
                )

            # Attempt to build an approximate sensor model from information in a corner coordinate TRE. The CSCRNA
            # TRE is considered more precise than IGEOLO so we will used whenever possible.
            if SensorModelTypes.PROJECTIVE in self.selected_sensor_model_types:
                cscrna_tre = self.xml_tres.find("./tre[@name='CSCRNA']")
                if cscrna_tre is not None:
                    approximate_sensor_model = ProjectiveSensorModelBuilder(
                        self.xml_tres, full_image_width, full_image_height
                    ).build()
                    if approximate_sensor_model is not None and chipped_image_info is not None:
                        approximate_sensor_model = ChippedImageSensorModel(
                            chipped_image_info.full_image_coordinates,
                            chipped_image_info.chipped_image_coordinates,
                            approximate_sensor_model,
                        )

                # TODO: Maybe create a projective sensor model from corner locations derived from the precision model
                # TODO: Consider using the rough corners from IGEOLO

        if self.xml_dess is not None and len(self.xml_dess) > 0:
            des_accessor = NITFDESAccessor(self.xml_dess)

            xml_data_content_segments = des_accessor.get_segments_by_name("XML_DATA_CONTENT")
            if xml_data_content_segments is not None:
                for xml_data_segment in xml_data_content_segments:
                    xml_bytes = des_accessor.parse_field_value(xml_data_segment, "DESDATA", base64.b64decode)
                    xml_str = xml_bytes.decode("utf-8")
                    if "SIDD" in xml_str:
                        # This looks like a SIDD file. Skip for now
                        # SIDD images will contain SICD extensions but the SIDD should come first
                        break
                    elif "SICD" in xml_str and SensorModelTypes.SICD in self.selected_sensor_model_types:
                        precision_sensor_model = SICDSensorModelBuilder(sicd_xml=xml_str).build()
                        break

        # If we have both an approximate and a precision sensor model return them as a composite so applications
        # can choose which model best meets their needs. If we were only able to construct one or the other then
        # return what we were able to build.
        if approximate_sensor_model is not None and precision_sensor_model is not None:
            return CompositeSensorModel(
                approximate_sensor_model=approximate_sensor_model,
                precision_sensor_model=precision_sensor_model,
            )
        elif precision_sensor_model is not None:
            return precision_sensor_model
        else:
            return approximate_sensor_model



from typing import Callable, List, TypeVar
from xml.etree import ElementTree as ET

# This is a type placeholder needed by the _get_tre_field_value() type hints
T = TypeVar("T")


def get_tre_field_value(tre: ET.Element, field_name: str, type_conversion: Callable[[str], T]) -> T:
    """
    This is a private utility function that will find a named "field" element in the children of a TRE Element and
    return the "value" attribute of that named field. A type conversion function can be provided to convert the
    attribute value to a specific Python type (e.g. int, float, or str)

    :param tre: the root element to find the named field in
    :param field_name: the name of the field element
    :param type_conversion: the desired type of the output, must support construction from a string

    :return: T = the value converted to the requested type
    """
    field_element = tre.find(f"./field[@name='{field_name}']")
    if field_element is None:
        raise ValueError(f"Unable to find TRE field named {field_name} in {tre.tag}")
    str_value = field_element.get("value")
    if str_value is None:
        raise ValueError(f"Field {field_name} does not have a value attribute.")
    return type_conversion(str_value)


def parse_rpc_coefficients(tre: ET.Element, repeated_name: str) -> List[float]:
    """
    This private utility function parses RPC coefficients from the child elements of a <repeated ...> tag in the
    XML TREs.

    :param tre: XML document
    :param repeated_name: find a specific name in the XML tree

    :return: a list of floating point values for the coefficients
    """

    repeated = tre.find(f"./repeated[@name='{repeated_name}']")
    if repeated is None:
        raise ValueError(f"XML TRE does not contain a repeated element named {repeated_name}")

    num_coefficients_str = repeated.get("number")
    if num_coefficients_str is None:
        raise ValueError("Invalid XML TRE: Repeated tag in XML TRE is missing required number attribute")
    num_coefficients = int(num_coefficients_str)

    result: List[float] = [0.0] * num_coefficients
    for group in repeated:
        index_str = group.get("index")
        if index_str is None:
            raise ValueError("Invalid XML TRE: Repeated group in XML TRE is missing required index attribute")
        index = int(index_str)

        if group[0] is None:
            raise ValueError("Invalid XML TRE: Repeated group in XML TRE is missing required field child element")

        value_str = group[0].get("value")
        if value_str is None:
            raise ValueError("Invalid XML TRE: Field in repeated group in XML TRE is missing required value attribute")
        value = float(value_str)
        result[index] = value

    return result



from abc import ABC, abstractmethod
from typing import Optional

from aws.osml.photogrammetry import SensorModel


class SensorModelBuilder(ABC):
    """
    This is an abstract base for all classes used to construct SensorModels from various types of metadata.
    """

    def __init__(self) -> None:
        """
        Constructor for the builder accepting various of required properties or format

        :return: None
        """
        pass

    @abstractmethod
    def build(self) -> Optional[SensorModel]:
        """
        Constructs the sensor model from the available information. Note that in cases where not enough information is
        available to provide any solution this method will return None.

        :return: the sensor model if available in the metadata provided
        """



import numpy as np
import numpy.typing as npt
import pyproj
from pyproj.enums import TransformDirection


class WorldCoordinate:
    """
    A world coordinate is a vector representing a position in 3D space. The ground coordinate system specified is
    either Geodetic (latitude, longitude, and height above the WGS 84 reference ellipsoid), or Rectangular (cartesian
    coordinates in reference to a local tangent plane). Regardless whether the coordinate system is specified as
    Geodetic or Rectangular, associated ground point locations are represented as a triple – x, y, and z.

    A Rectangular system should be specified when the image footprint is near the earth’s North or South Pole. Either
    a Rectangular or Geodetic system can be specified when the footprint is near 180 degrees East longitude. However,
    if Geodetic, the range for longitude is then specified as (0,2pi) radians instead of the usual (-pi, +pi) radians.
    """

    def __init__(self, coordinate: npt.ArrayLike = None) -> None:
        """
        Constructs a world coordinate from an x, y, z triple. The triple can be expressed as a List or any other
        structure that can be used to construct a Numpy array.

        :param coordinate: the x,y,z components

        :return: None
        """
        if coordinate is None:
            coordinate = [0.0, 0.0, 0.0]

        if len(coordinate) != 3:
            raise ValueError("WorldCoordinates must have 3 components (x,y,z)")

        self.coordinate = np.array(coordinate, dtype=np.float64)

    @property
    def x(self) -> float:
        return self.coordinate[0]

    @x.setter
    def x(self, value: float) -> None:
        self.coordinate[0] = value

    @property
    def y(self) -> float:
        return self.coordinate[1]

    @y.setter
    def y(self, value: float) -> None:
        self.coordinate[1] = value

    @property
    def z(self) -> float:
        return self.coordinate[2]

    @z.setter
    def z(self, value: float) -> None:
        self.coordinate[2] = value


class GeodeticWorldCoordinate(WorldCoordinate):
    """
    A GeodeticWorldCoordinate is an WorldCoordinate where the x,y,z components can be interpreted as longitude,
    latitude, and elevation.
    """

    def __init__(self, coordinate: npt.ArrayLike = None) -> None:
        """
        Constructs a geodetic world coordinate from a longitude, latitude, elevation triple. The longitude and
        latitude components are in radians. The triple can be expressed as a List or any other structure that can
        be used to construct a Numpy array.

        :param coordinate: the longitude, latitude, elevation components

        :return: None
        """
        super().__init__(coordinate)

    @property
    def longitude(self) -> float:
        return self.x

    @longitude.setter
    def longitude(self, value: float) -> None:
        self.x = value

    @property
    def latitude(self) -> float:
        return self.y

    @latitude.setter
    def latitude(self, value: float) -> None:
        self.y = value

    @property
    def elevation(self) -> float:
        return self.z

    @elevation.setter
    def elevation(self, value: float) -> None:
        self.z = value

    def to_dms_string(self) -> str:
        """
        Outputs this coordinate in a format ddmmssXdddmmssY. The first half, ddmmssX, represents degrees, minutes, and
        seconds of latitude with X representing North or South (N for North, S for South). The second half, dddmmssY,
        represents degrees, minutes, and seconds of longitude with Y representing East or West (E for East, W for West),
        respectively.

        :return: the formatted coordinate string
        """
        result_parts = []
        lat_degrees = np.degrees(self.latitude)
        direction = "N"
        if lat_degrees < 0:
            lat_degrees *= -1.0
            direction = "S"
        d = int(lat_degrees)
        m = int(round(lat_degrees - d, 6) * 60)
        s = int(round(lat_degrees - d - m / 60, 6) * 3600)
        result_parts.append(format(d, "02d"))
        result_parts.append(format(m, "02d"))
        result_parts.append(format(s, "02d"))
        result_parts.append(direction)

        lon_degrees = np.degrees(self.longitude)
        direction = "E"
        if lon_degrees < 0:
            lon_degrees *= -1.0
            direction = "W"
        d = int(lon_degrees)
        m = int(round(lon_degrees - d, 6) * 60)
        s = int(round(lon_degrees - d - m / 60, 6) * 3600)
        result_parts.append(format(d, "03d"))
        result_parts.append(format(m, "02d"))
        result_parts.append(format(s, "02d"))
        result_parts.append(direction)

        return "".join(result_parts)


# These are common definitions of projections used by Pyproj. They are used when converting between an Earth Centered
# Earth Fixed (ECEF or geocentric) coordinate system that uses cartesian coordinates and a longitude, latitude based
# geographic coordinate system. Both of these systems use the WGS84 datum which is a widely used standard among our
# customers
ECEF_PROJ = pyproj.Proj(proj="geocent", ellps="WGS84", datum="WGS84")
LLA_PROJ = pyproj.Proj(proj="latlong", ellps="WGS84", datum="WGS84")
GEODETIC_TO_GEOCENTRIC_TRANSFORM = pyproj.Transformer.from_proj(LLA_PROJ, ECEF_PROJ)


def geocentric_to_geodetic(ecef_world_coordinate: WorldCoordinate) -> GeodeticWorldCoordinate:
    """
    Converts a ECEF world coordinate (x, y, z) in meters into a (longitude, latitude, elevation) geodetic coordinate
    with units of radians, radians, meters.

    :param ecef_world_coordinate: the geocentric coordinate

    :return: the geodetic coordinate
    """
    return GeodeticWorldCoordinate(
        GEODETIC_TO_GEOCENTRIC_TRANSFORM.transform(
            ecef_world_coordinate.x,
            ecef_world_coordinate.y,
            ecef_world_coordinate.z,
            radians=True,
            direction=TransformDirection.INVERSE,
        )
    )


def geodetic_to_geocentric(geodetic_coordinate: GeodeticWorldCoordinate) -> WorldCoordinate:
    """
    Converts a geodetic world coordinate (longitude, latitude, elevation) with units of radians, radians, meters into
    a ECEF / geocentric world coordinate (x, y, z) in meters.

    :param geodetic_coordinate: the geodetic coordinate

    :return: the geocentric coordinate
    """
    return WorldCoordinate(
        GEODETIC_TO_GEOCENTRIC_TRANSFORM.transform(
            geodetic_coordinate.longitude,
            geodetic_coordinate.latitude,
            geodetic_coordinate.elevation,
            radians=True,
            direction=TransformDirection.FORWARD,
        )
    )


class ImageCoordinate:
    """
    This image coordinate system convention is defined as follows. The upper left corner of the upper left pixel
    of the original full image has continuous image coordinates (pixel position) (r, c) = (0.0,0.0) , and the center
    of the upper left pixel has continuous image coordinates (r, c) = (0.5,0.5) . The first row of the original full
    image has discrete image row coordinate R = 0 , and corresponds to a range of continuous image row coordinates of
    r = [0,1) . The first column of the original full image has discrete image column coordinate C = 0 , and
    corresponds to a range of continuous image column coordinates of c = [0,1) . Thus, for example, continuous image
    coordinates (r, c) = (5.6,8.3) correspond to the sixth row and ninth column of the original full image, and
    discrete image coordinates (R,C) = (5,8).
    """

    def __init__(self, coordinate: npt.ArrayLike = None) -> None:
        """
        Constructs an image coordinate from an x, y tuple. The tuple can be expressed as a List or any other
        structure that can be used to construct a Numpy array.

        :param coordinate: the x, y components

        :return: None
        """
        if coordinate is None:
            coordinate = [0.0, 0.0]

        if len(coordinate) != 2:
            raise ValueError("ImageCoordinates must have 2 components (x,y)")

        self.coordinate = np.array(coordinate, dtype=np.float64)

    @property
    def c(self) -> float:
        return self.coordinate[0]

    @c.setter
    def c(self, value: float) -> None:
        self.coordinate[0] = value

    @property
    def r(self) -> float:
        return self.coordinate[1]

    @r.setter
    def r(self, value: float) -> None:
        self.coordinate[1] = value

    @property
    def x(self) -> float:
        return self.c

    @x.setter
    def x(self, value: float) -> None:
        self.c = value

    @property
    def y(self) -> float:
        return self.r

    @y.setter
    def y(self, value: float) -> None:
        self.r = value



import base64
import logging
from secrets import token_hex
from typing import Any, Dict, List, Optional

from defusedxml import ElementTree
from osgeo import gdal, gdalconst

from aws.osml.gdal import GDALCompressionOptions, GDALImageFormats, NITFDESAccessor, RangeAdjustmentType, get_type_and_scales
from aws.osml.photogrammetry import ImageCoordinate, SensorModel

from .sicd_updater import SICDUpdater

logger = logging.getLogger(__name__)


class GDALTileFactory:
    """
    This class creates tiles from a larger image on request. Image metadata is retained whenever possible but updated
    as necessary to account for the new raster bounds.
    """

    def __init__(
        self,
        raster_dataset: gdal.Dataset,
        sensor_model: Optional[SensorModel] = None,
        tile_format: GDALImageFormats = GDALImageFormats.NITF,
        tile_compression: GDALCompressionOptions = GDALCompressionOptions.NONE,
        output_type: Optional[int] = None,
        range_adjustment: RangeAdjustmentType = RangeAdjustmentType.NONE,
    ):
        """
        Constructs a new factory capable of producing tiles from a given GDAL raster dataset.

        :param raster_dataset: the original raster dataset to create tiles from
        :param sensor_model: the sensor model providing mensuration support for this image
        :param tile_format: the output tile format
        :param tile_compression: the output tile compression
        :param output_type: the GDAL pixel type in the output tile
        :param range_adjustment: the type of scaling used to convert raw pixel values to the output range
        """
        self.tile_format = tile_format
        self.tile_compression = tile_compression
        self.raster_dataset = raster_dataset
        self.sensor_model = sensor_model
        self.des_accessor = None
        self.sicd_updater = None
        self.sicd_des_header = None
        self.range_adjustment = range_adjustment
        self.output_type = output_type

        if self.raster_dataset.GetDriver().ShortName == "NITF":
            xml_des = self.raster_dataset.GetMetadata("xml:DES")
            self.des_accessor = NITFDESAccessor(xml_des)

            xml_data_content_segments = self.des_accessor.get_segments_by_name("XML_DATA_CONTENT")
            if xml_data_content_segments is not None and len(xml_data_content_segments) > 0:
                # This appears to be SICD or SIDD data
                # TODO: Check to make sure this is the right XML_DATA_CONTENT segment containing SICD
                sicd_des = xml_data_content_segments[0]
                sicd_bytes = self.des_accessor.parse_field_value(sicd_des, "DESDATA", base64.b64decode)
                sicd_xml = sicd_bytes.decode("utf-8")
                sicd_metadata = ElementTree.fromstring(sicd_xml)
                self.sicd_des_header = self.des_accessor.extract_des_header(sicd_des)
                self.sicd_updater = SICDUpdater(sicd_metadata)

        self.default_gdal_translate_kwargs = self._create_gdal_translate_kwargs()

    def create_encoded_tile(self, src_window: List[int]) -> Optional[bytearray]:
        """
        This method cuts a tile from the full image, updates the metadata as needed, and finally compresses/encodes
        the result in the output format requested.

        :param src_window: the [left_x, top_y, width, height] bounds of this tile
        :return: the encoded image tile or None if one could not be produced
        """
        temp_ds_name = f"/vsimem/{token_hex(16)}.{self.tile_format}"

        # Use the request and metadata from the raster dataset to create a set of keyword
        # arguments for the gdal.Translate() function. This will configure that function to
        # create image tiles using the format, compression, etc. requested by the client.
        gdal_translate_kwargs = self.default_gdal_translate_kwargs.copy()

        # Create a new IGEOLO value based on the corner points of this tile
        if self.sensor_model is not None and self.tile_format == GDALImageFormats.NITF:
            gdal_translate_kwargs["creationOptions"].append("ICORDS=G")
            gdal_translate_kwargs["creationOptions"].append("IGEOLO=" + self.create_new_igeolo(src_window))

        if self.sicd_updater is not None and self.tile_format == GDALImageFormats.NITF:
            # If we're outputting a SICD tile we need to update the XML metadata to include the new chip
            # origin and size. This will allow applications using the tile to correctly interpret the remaining
            # image metadata.
            self.sicd_updater.update_image_data_for_chip(src_window)
            updated_sicd_des = self.sicd_des_header + self.sicd_updater.encode_current_xml()

            gdal_translate_kwargs["creationOptions"].append("ICAT=SAR")
            gdal_translate_kwargs["creationOptions"].append("IREP=NODISPLY")
            gdal_translate_kwargs["creationOptions"].append("IREPBAND= , ")
            gdal_translate_kwargs["creationOptions"].append("ISUBCAT=I,Q")
            gdal_translate_kwargs["creationOptions"].append("DES=XML_DATA_CONTENT=" + updated_sicd_des)

        # Use GDAL to create an encoded tile of the image region
        # From GDAL documentation:
        #   srcWin --- subwindow in pixels to extract:
        #               [left_x, top_y, width, height]
        gdal.Translate(
            temp_ds_name,
            self.raster_dataset,
            srcWin=src_window,
            **gdal_translate_kwargs,
        )

        # Read the VSIFile
        vsifile_handle = gdal.VSIFOpenL(temp_ds_name, "r")
        if vsifile_handle is None:
            return None
        stat = gdal.VSIStatL(temp_ds_name, gdal.VSI_STAT_SIZE_FLAG)
        vsibuf = gdal.VSIFReadL(1, stat.size, vsifile_handle)
        return vsibuf

    def create_new_igeolo(self, src_window: List[int]) -> str:
        """
        Create a new 60 character string representing the corner coordinates of this tile. The string conforms to
        the geographic "G" choice for the ICORDS field as specified in MIL-STD-2500C Table A-3 NITF Image Subheader.

        The locations of the four corners of are encoded image coordinate order:
        (0,0), (Max X, 0), (Max X, Max Y), (0, Max Y) with each corner represented as a ddmmssXdddmmssY string
        representing latitude and longitude. The first half, ddmmssX, represents degrees, minutes, and seconds of
        latitude with X representing North or South (N for North, S for South). The second half, dddmmssY, represents
        degrees, minutes, and seconds of longitude with Y representing East or West (E for East, W for West),
        respectively.

        :param src_window: the [left_x, top_y, width, height] bounds of this tile
        :return: the 60 character IGEOLO geographic coordinate string
        """
        tile_corners = [
            [src_window[0], src_window[1]],
            [src_window[0] + src_window[2], src_window[1]],
            [src_window[0] + src_window[2], src_window[1] + src_window[3]],
            [src_window[0], src_window[1] + src_window[3]],
        ]
        dms_coords = []
        for coord in tile_corners:
            geodetic_tile_corner = self.sensor_model.image_to_world(ImageCoordinate(coord))
            dms_coords.append(geodetic_tile_corner.to_dms_string())
        return "".join(dms_coords)

    def _create_gdal_translate_kwargs(self) -> Dict[str, Any]:
        """
        This method creates a set of keyword arguments suitable for passing to the gdal.Translate
        function. The values for these options are derived from the region processing request and
        the raster dataset itself.

        See: https://gdal.org/python/osgeo.gdal-module.html#Translate
        See: https://gdal.org/python/osgeo.gdal-module.html#TranslateOptions

        :return: Dict[str, any] = the dictionary of translate keyword arguments
        """
        # Figure out what type of image this is and calculate a scale to map input pixels to the output type
        output_type, scale_params = get_type_and_scales(
            self.raster_dataset, desired_output_type=self.output_type, range_adjustment=self.range_adjustment
        )

        gdal_translate_kwargs = {
            "scaleParams": scale_params,
            "outputType": output_type,
            "format": self.tile_format,
        }

        creation_options = []
        if self.tile_format == GDALImageFormats.NITF:
            # Creation options specific to the NITF raster driver.
            # See: https://gdal.org/drivers/raster/nitf.html
            if self.tile_compression is None or self.tile_compression == GDALCompressionOptions.J2K:
                # Default NITF tiles to JPEG2000 compression if not specified
                # The OpenJPEG library does not currently support floating point pixel types. If J2K compression
                # was requested we need to fall back to no compression to avoid a failure. This update can be
                # removed if compression of floating points is supported or if we move to an alternate J2K library.
                if output_type in [gdalconst.GDT_Float32, gdalconst.GDT_Float64]:
                    logger.warning(
                        "OpenJPEG J2K library does not support floating point pixel values. "
                        "Outputs will not be compressed."
                    )
                    creation_options.append("IC=NC")
                else:
                    creation_options.append("IC=C8")
            elif self.tile_compression == GDALCompressionOptions.JPEG:
                creation_options.append("IC=C3")
            elif self.tile_compression == GDALCompressionOptions.NONE:
                creation_options.append("IC=NC")
            else:
                logging.warning("Invalid compress specified for NITF image defaulting to JPEG2000!")
                creation_options.append("IC=C8")
        elif self.tile_format == GDALImageFormats.GTIFF:
            # Creation options specific to the GeoTIFF raster driver.
            # See: https://gdal.org/drivers/raster/nitf.html
            if self.tile_compression is None:
                # Default GeoTiff tiles to LZQ compression if not specified
                creation_options.append("COMPRESS=LZW")
            elif self.tile_compression == GDALCompressionOptions.LZW:
                creation_options.append("COMPRESS=LZW")
            elif self.tile_compression == GDALCompressionOptions.JPEG:
                creation_options.append("COMPRESS=JPEG")
            elif self.tile_compression == GDALCompressionOptions.NONE:
                creation_options.append("COMPRESS=NONE")
            else:
                logging.warning("Invalid compress specified for GTIFF image defaulting to LZW!")
                creation_options.append("COMPRESS=LZW")

        gdal_translate_kwargs["creationOptions"] = creation_options

        return gdal_translate_kwargs



from math import degrees, radians, sqrt
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from scipy.optimize import minimize

from . import ConstantElevationModel, ElevationModel, GeodeticWorldCoordinate, ImageCoordinate, WorldCoordinate
from .math_utils import equilateral_triangle
from .sensor_model import SensorModel, SensorModelOptions


class RPCPolynomial:
    def __init__(self, coefficients: List[float]) -> None:
        """
        The order of terms in the coefficients is defined to match the RPC00B specification as defined in STDI-0002
        Volume 1 Appendix E Section E.2.4. Note that the order of terms for coefficients taken from other TREs (e.g.
        RPC00A is different).

        :param coefficients: the polynomial coefficients

        :return: None
        """
        self.coefficients = coefficients

    def evaluate(self, normalized_world_coordinate: WorldCoordinate) -> float:
        """
        This function evaluates the polynomial for the given world coordinate by summing up the result of applying
        each coefficient to the world coordinate components. Note that these polynomials are usually defined with
        the assumption that the world coordinate has been normalized.

        :param normalized_world_coordinate: the world coordinate

        :return: the resulting value
        """
        # The equations in the RPC00B definition use L, P, H to denote the normalized longitude, latitude, and
        # elevation components respectively.
        l_val = normalized_world_coordinate.x
        p_val = normalized_world_coordinate.y
        h_val = normalized_world_coordinate.z

        result = 0.0
        result += self.coefficients[0]  # constant
        result += self.coefficients[1] * l_val  # L
        result += self.coefficients[2] * p_val  # P
        result += self.coefficients[3] * h_val  # H
        result += self.coefficients[4] * l_val * p_val  # LP
        result += self.coefficients[5] * l_val * h_val  # LH
        result += self.coefficients[6] * p_val * h_val  # PH
        result += self.coefficients[7] * l_val * l_val  # LL
        result += self.coefficients[8] * p_val * p_val  # PP
        result += self.coefficients[9] * h_val * h_val  # HH
        result += self.coefficients[10] * l_val * p_val * h_val  # LPH
        result += self.coefficients[11] * l_val * l_val * l_val  # LLL
        result += self.coefficients[12] * l_val * p_val * p_val  # LPP
        result += self.coefficients[13] * l_val * h_val * h_val  # LHH
        result += self.coefficients[14] * l_val * l_val * p_val  # LLP
        result += self.coefficients[15] * p_val * p_val * p_val  # PPP
        result += self.coefficients[16] * p_val * h_val * h_val  # PHH
        result += self.coefficients[17] * l_val * l_val * h_val  # LLH
        result += self.coefficients[18] * p_val * p_val * h_val  # PPH
        result += self.coefficients[19] * h_val * h_val * h_val  # HHH
        return result

    def __call__(self, *args, **kwargs) -> float:
        """
        This makes the polynomial object callable such that it can be applied to a world coordinate directly e.g.
        result = polynomial(coordinate).

        :param args: the world coordinate
        :param kwargs: no keyword arguments are accepted but this is required for the Callable interface
        :return: the result of the evaluation
        """
        return self.evaluate(args[0])


class RPCSensorModel(SensorModel):
    """
    A Rational Polynomial Camera (RPC) sensor model is one where the world to image transform is approximated using
    a ratio of polynomials. The polynomials capture specific relationships between the latitude, longitude, and
    elevation and the image pixels.

    These cameras were common approximations for many years but started to be phased out in 2014 in favor of the
    more general Replacement Sensor Model (RSM). It is not uncommon to find historical imagery or imagery from
    older sensors still operating today that contain the metadata for these sensor models.
    """

    def __init__(
        self,
        err_bias: float,
        err_rand: float,
        line_off: float,
        samp_off: float,
        lat_off: float,
        long_off: float,
        height_off: float,
        line_scale: float,
        samp_scale: float,
        lat_scale: float,
        long_scale: float,
        height_scale: float,
        line_num_poly: RPCPolynomial,
        line_den_poly: RPCPolynomial,
        samp_num_poly: RPCPolynomial,
        samp_den_poly: RPCPolynomial,
    ) -> None:
        """
        This constructs the sensor model using parameters normally found in the NITF RPC00B TRE. Note that there also
        exists RPC00A, and RPC00C tags that contain the same information. Be careful and check the specifications. The
        order of the polynomial coefficients is different even though the structure of the TRE is the same.

        :param err_bias: non time-varying error estimate assumes correlated images
        :param err_rand: time-varying error estimate assumes uncorrelated images
        :param line_off: offset used to normalize/denormalize image y components
        :param samp_off: offset used to normalize/denormalize image x components
        :param lat_off: offset used to normalize/denormalize world y components
        :param long_off: offset used to normalize/denormalize world x components
        :param height_off: offset used to normalize/denormalize world z components
        :param line_scale: scale used to normalize/denormalize image y components
        :param samp_scale: scale used to normalize/denormalize image x components
        :param lat_scale: scale used to normalize/denormalize world y components
        :param long_scale: scale used to normalize/denormalize world x components
        :param height_scale: scale used to normalize/denormalize world z components
        :param line_num_poly: polynomial used as the numerator in line (row) calculations
        :param line_den_poly: polynomial used as the denominator in line (row) calculations
        :param samp_num_poly: polynomial used as the numerator in sample (column) calculations
        :param samp_den_poly: polynomial used as the denominator in sample (column) calculations

        :return: None
        """
        super().__init__()
        self.err_bias = err_bias
        self.err_rand = err_rand
        self.line_off = line_off
        self.samp_off = samp_off
        self.lat_off = lat_off
        self.long_off = long_off
        self.height_off = height_off
        self.line_scale = line_scale
        self.samp_scale = samp_scale
        self.lat_scale = lat_scale
        self.long_scale = long_scale
        self.height_scale = height_scale
        self.line_numerator_poly = line_num_poly
        self.line_denominator_poly = line_den_poly
        self.samp_numerator_poly = samp_num_poly
        self.samp_denominator_poly = samp_den_poly
        self.default_elevation_model = ConstantElevationModel(self.height_off)

    def world_to_image(self, geodetic_coordinate: GeodeticWorldCoordinate) -> ImageCoordinate:
        """
        This function transforms a geodetic world coordinate (longitude, latitude, elevation) into an image coordinate
        (x, y).

        :param geodetic_coordinate: the world coordinate (longitude, latitude, elevation)

        :return: the resulting image coordinate (x,y)
        """
        # The RPC spec assumes the ground coordinates are latitude and longitude in units of decimal degrees and
        # the geodetic elevation in units of meters. The ground coordinates are referenced to WGS-84. The
        # GeodeticWorldCoordinate is assumed to be in radians with meters elevation. Convert the coordinate to
        # degrees before normalizing the values.
        # TODO: Check that GeodeticWorldCoordinate in RSM is meters elevation
        l_val = (degrees(geodetic_coordinate.x) - self.long_off) / self.long_scale
        p_val = (degrees(geodetic_coordinate.y) - self.lat_off) / self.lat_scale
        h_val = (geodetic_coordinate.z - self.height_off) / self.height_scale
        norm_domain_coordinate = WorldCoordinate([l_val, p_val, h_val])

        # Evaluate the rational polynomials
        cn = self.samp_numerator_poly(norm_domain_coordinate) / self.samp_denominator_poly(norm_domain_coordinate)
        rn = self.line_numerator_poly(norm_domain_coordinate) / self.line_denominator_poly(norm_domain_coordinate)

        # Denormalize the rational polynomial results to get the actual x,y of the image coordinate
        col = cn * self.samp_scale + self.samp_off
        row = rn * self.line_scale + self.line_off
        return ImageCoordinate([col, row])

    def image_to_world(
        self,
        image_coordinate: ImageCoordinate,
        elevation_model: Optional[ElevationModel] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> GeodeticWorldCoordinate:
        """
        This function implements the image to world transform by iteratively invoking world to image within a
        minimization routine to find a matching image coordinate. The longitude and latitude parameters are searched
        independently while the elevation of the world coordinate comes from the elevation model.

        :param image_coordinate: the image coordinate (x, y)
        :param elevation_model: an optional elevation model used to transform the coordinate
        :param options: optional hints, supports initial_guess and initial_search_distance

        :return: the corresponding world coordinate
        """

        # This is the function we will be minimizing. Given an x,y coordinate in the ground domain we use invoke the
        # ground_domain_to_image function to get a projection of that location in the image. Then we compute the
        # distance between that new image location and the input image location. When those locations match then
        # we know we have the ground domain coordinate that corresponds to the input.
        def distance_to_target_coordinate(lonlat_coord: Tuple[float, float]) -> float:
            ground_domain_coordinate = GeodeticWorldCoordinate([lonlat_coord[0], lonlat_coord[1], 0.0])
            self.default_elevation_model.set_elevation(ground_domain_coordinate)
            if elevation_model:
                elevation_model.set_elevation(ground_domain_coordinate)
            new_image_coordinate = self.world_to_image(ground_domain_coordinate)
            return sqrt(
                (image_coordinate.x - new_image_coordinate.x) ** 2 + (image_coordinate.y - new_image_coordinate.y) ** 2
            )

        # Select an initial guess using the normalization offsets for this camera model. Normally these are values
        # near an image corner or the center
        initial_guess = options.get(SensorModelOptions.INITIAL_GUESS) if options is not None else None
        if initial_guess is None:
            initial_guess = np.array([radians(self.long_off), radians(self.lat_off)])
        if isinstance(initial_guess, List):
            initial_guess = np.array(initial_guess)

        initial_search_distance = options.get(SensorModelOptions.INITIAL_SEARCH_DISTANCE) if options is not None else None
        if initial_search_distance is None:
            initial_search_distance = radians(0.5)

        # Iteratively adjust the initial guess to minimize the distance to the target image coordinate. We are only
        # allowing the x,y components to vary here and the z is fixed to the elevation model. The starting simplex
        # is estimated as a triangle centered on the normalization offsets for this RPC.
        res = minimize(
            distance_to_target_coordinate,
            initial_guess,
            method="Nelder-Mead",
            options={
                "xatol": radians(0.000001),
                "fatol": 0.5,
                "initial_simplex": equilateral_triangle(initial_guess.tolist(), initial_search_distance),
            },
        )

        # The minimization result is an (x,y) tuple, so we need to expand it to x,y,z and replace the z component with
        # the height from the elevation model. Note that the units of this are radians, radians, meters
        world_coordinate = GeodeticWorldCoordinate(np.append(res.x, 0.0))
        self.default_elevation_model.set_elevation(world_coordinate)
        if elevation_model:
            elevation_model.set_elevation(world_coordinate)

        return world_coordinate



import logging
from typing import Any, Optional, Tuple

import numpy as np
from osgeo import gdal

from aws.osml.photogrammetry import (
    DigitalElevationModelTileFactory,
    ElevationRegionSummary,
    GDALAffineSensorModel,
    ImageCoordinate,
    geodetic_to_geocentric,
)


class GDALDigitalElevationModelTileFactory(DigitalElevationModelTileFactory):
    """
    This tile factory uses GDAL to load elevation data into numpy arrays. Any raster format supported by GDAL is
    fair game but the format must have sufficient metadata to populate the GDAL geo transform.
    """

    def __init__(self, tile_directory: str) -> None:
        """
        Constructor for the factory that takes in the root location of the elevation tiles. If the root starts with
        s3:/ then GDAL's VSIS3 virtual file system will be used to read rasters directly from cloud storage.

        :param tile_directory: the root tile location, may be an S3 URL

        :return: None
        """
        super().__init__()
        self.tile_directory = tile_directory

    def get_tile(
        self, tile_path: str
    ) -> Tuple[Optional[Any], Optional[GDALAffineSensorModel], Optional[ElevationRegionSummary]]:
        """
        Retrieve a numpy array of elevation values and a sensor model.

        TODO: Replace Any with numpy.typing.ArrayLike once we move to numpy >1.20

        :param tile_path: the location of the tile to load

        :return: an array of elevation values, a sensor model, and a summary or (None, None, None)
        """
        tile_location = f"{self.tile_directory}/{tile_path}"
        tile_location = tile_location.replace("s3:/", "/vsis3", 1)
        ds = gdal.Open(tile_location)

        # It isn't unusual for a DEM tile set to be missing tiles for regions (particularly over the ocean). If
        # the raster dataset can't be opened we'll return nothing so the client can proceed knowing that the
        # information isn't available.
        if not ds:
            logging.debug(f"No DEM tile available for {tile_path}. Checked {tile_location}")
            return None, None, None

        # If the raster exists but doesn't have a geo transform then it is likely invalid input data.
        geo_transform = ds.GetGeoTransform(can_return_null=True)
        if not geo_transform:
            logging.warning(f"DEM tile does not have geo transform metadata and can't be used: {tile_location}")
            return None, None, None

        raster_band = ds.GetRasterBand(1)
        band_as_array = raster_band.ReadAsArray(0, 0, ds.RasterXSize, ds.RasterYSize)
        height, width = band_as_array.shape
        sensor_model = GDALAffineSensorModel(geo_transform)

        # Compute the distance in meters from the upper left to the lower right corner. Divide that by the distance
        # in pixels to get an approximate pixel size in meters
        ul_corner_ecf = geodetic_to_geocentric(sensor_model.image_to_world(ImageCoordinate([0, 0]))).coordinate
        lr_corner_ecf = geodetic_to_geocentric(sensor_model.image_to_world(ImageCoordinate([width, height]))).coordinate
        post_spacing = np.linalg.norm(ul_corner_ecf - lr_corner_ecf) / np.sqrt(width * width + height * height)

        stats = raster_band.GetStatistics(True, True)
        summary = ElevationRegionSummary(
            min_elevation=stats[0],
            max_elevation=stats[1],
            no_data_value=raster_band.GetNoDataValue(),
            post_spacing=post_spacing,
        )

        return band_as_array, sensor_model, summary



# Telling flake8 to not flag errors in this file. It is normal that these classes are imported but not used in an
# __init__.py file.
# flake8: noqa
"""
The photogrammetry package contains implementations of various sensor and elevation models used to convert between
the image (x, y) and geodetic (lon, lat, elev) coordinate systems.
"""

from .chipped_image_sensor_model import ChippedImageSensorModel
from .composite_sensor_model import CompositeSensorModel
from .coordinates import (
    GeodeticWorldCoordinate,
    ImageCoordinate,
    WorldCoordinate,
    geocentric_to_geodetic,
    geodetic_to_geocentric,
)
from .digital_elevation_model import DigitalElevationModel, DigitalElevationModelTileFactory, DigitalElevationModelTileSet
from .elevation_model import ConstantElevationModel, ElevationModel, ElevationRegionSummary
from .gdal_sensor_model import GDALAffineSensorModel
from .projective_sensor_model import ProjectiveSensorModel
from .replacement_sensor_model import (
    RSMContext,
    RSMGroundDomain,
    RSMGroundDomainForm,
    RSMImageDomain,
    RSMLowOrderPolynomial,
    RSMPolynomial,
    RSMPolynomialSensorModel,
    RSMSectionedPolynomialSensorModel,
)
from .rpc_sensor_model import RPCPolynomial, RPCSensorModel
from .sensor_model import SensorModel, SensorModelOptions
from .sicd_sensor_model import (
    COAProjectionSet,
    INCAProjectionSet,
    PFAProjectionSet,
    PlaneProjectionSet,
    Polynomial2D,
    PolynomialXYZ,
    RGAZCOMPProjectionSet,
    SARImageCoordConverter,
    SICDSensorModel,
)
from .srtm_dem_tile_set import SRTMTileSet

__all__ = [
    "ChippedImageSensorModel",
    "CompositeSensorModel",
    "GeodeticWorldCoordinate",
    "ImageCoordinate",
    "WorldCoordinate",
    "geocentric_to_geodetic",
    "geodetic_to_geocentric",
    "DigitalElevationModel",
    "DigitalElevationModelTileFactory",
    "DigitalElevationModelTileSet",
    "ConstantElevationModel",
    "ElevationModel",
    "ElevationRegionSummary",
    "GDALAffineSensorModel",
    "SARImageCoordConverter",
    "INCAProjectionSet",
    "PlaneProjectionSet",
    "PFAProjectionSet",
    "PolynomialXYZ",
    "Polynomial2D",
    "ProjectiveSensorModel",
    "RGAZCOMPProjectionSet",
    "RSMContext",
    "RSMGroundDomain",
    "RSMGroundDomainForm",
    "RSMImageDomain",
    "RSMLowOrderPolynomial",
    "RSMPolynomial",
    "RSMPolynomialSensorModel",
    "RSMSectionedPolynomialSensorModel",
    "RPCPolynomial",
    "RPCSensorModel",
    "SensorModel",
    "SensorModelOptions",
    "SICDSensorModel",
    "SRTMTileSet",
]



import logging
from typing import Optional, Union

import numpy as np
from xsdata.formats.dataclass.parsers import XmlParser

import aws.osml.formats.sicd.models.sicd_v1_2_1 as sicd121
import aws.osml.formats.sicd.models.sicd_v1_3_0 as sicd130

from ..photogrammetry import (
    ImageCoordinate,
    INCAProjectionSet,
    PFAProjectionSet,
    PlaneProjectionSet,
    Polynomial2D,
    PolynomialXYZ,
    RGAZCOMPProjectionSet,
    SARImageCoordConverter,
    SensorModel,
    SICDSensorModel,
    WorldCoordinate,
)
from .sensor_model_builder import SensorModelBuilder

logger = logging.getLogger(__name__)


def xyztype_to_ndarray(xyztype: Union[sicd121.XYZType, sicd130.XYZType]) -> np.ndarray:
    """
    Convert the XYZType to a 1-d NumPy array.

    :param xyztype: the XYZType dataclass
    :return: the NumPy array
    """
    return np.array([xyztype.x, xyztype.y, xyztype.z])


def poly1d_to_native(poly1d: Union[sicd121.Poly1DType, sicd130.XYZType]) -> np.polynomial.Polynomial:
    """
    Convert the Poly1DType to a NumPy Polynomial.

    :param poly1d: the Poly1D dataclass
    :return: the NumPy polynomial with matching coefficients
    """
    coefficients = [0] * (poly1d.order1 + 1)
    for coef in poly1d.coef:
        coefficients[coef.exponent1] = coef.value
    return np.polynomial.Polynomial(coefficients)


def poly2d_to_native(poly2d: Union[sicd121.Poly2DType, sicd130.Poly2DType]) -> Polynomial2D:
    """
    Convert the Poly2D dataclass to a Polynomial2D.

    :param poly2d: the Poly2D dataclass
    :return: the Polynomial2D with matching coefficients
    """
    coefficients = [0] * (poly2d.order1 + 1)
    for row in range(0, len(coefficients)):
        coefficients[row] = [0] * (poly2d.order2 + 1)
    for coef in poly2d.coef:
        coefficients[coef.exponent1][coef.exponent2] = coef.value
    return Polynomial2D(np.array(coefficients))


def xyzpoly_to_native(xyzpoly: Union[sicd121.XYZPolyType, sicd130.XYZPolyType]) -> PolynomialXYZ:
    """
    Convert the XYZPoly dataclass into a PolynomialXYZ.

    :param xyzpoly: the XYZPoly dataclass
    :return: the PolynomialXYZ with matching coefficients
    """
    return PolynomialXYZ(
        x_polynomial=poly1d_to_native(xyzpoly.x),
        y_polynomial=poly1d_to_native(xyzpoly.y),
        z_polynomial=poly1d_to_native(xyzpoly.z),
    )


class SICDSensorModelBuilder(SensorModelBuilder):
    """
    This builder is used to create sensor models for images that have SICD metadata. The metadata is provided
    as XML that conforms to the SICD specifications. We intend to support multiple SICD versions but the current
    software was implemented using the v1.2.1 and v1.3.0 specifications.
    """

    def __init__(self, sicd_xml: str):
        """
        Construct the builder given the SICD XML.

        :param sicd_xml: the XML string
        """
        super().__init__()
        self.sicd_xml = sicd_xml

    def build(self) -> Optional[SensorModel]:
        """
        Attempt to build a precise SAR sensor model. This sensor model handles chipped images natively.

        :return: the sensor model; if available
        """
        try:
            if self.sicd_xml is None or len(self.sicd_xml) == 0:
                return None

            parser = XmlParser()
            sicd = parser.from_string(self.sicd_xml)
            return SICDSensorModelBuilder.from_dataclass(sicd)
        except Exception as e:
            logging.error("Exception caught attempting to build SICD sensor model.", e)
        return None

    @staticmethod
    def from_dataclass(sicd: Union[sicd121.SICD, sicd130.SICD]) -> Optional[SensorModel]:
        """
        This method constructs a SICD sensor model from the python dataclasses generated when parsing the XML.

        :param sicd: the dataclass object constructed from the XML
        :return: the sensor model; if available
        """

        scp_ecf = WorldCoordinate(xyztype_to_ndarray(sicd.geo_data.scp.ecf))
        scp_pixel = ImageCoordinate([sicd.image_data.scppixel.col, sicd.image_data.scppixel.row])
        time_coa_poly = poly2d_to_native(sicd.grid.time_coapoly)
        arp_poly = xyzpoly_to_native(sicd.position.arppoly)

        coord_converter = SARImageCoordConverter(
            scp_pixel=scp_pixel,
            scp_ecf=scp_ecf,
            u_row=xyztype_to_ndarray(sicd.grid.row.uvect_ecf),
            u_col=xyztype_to_ndarray(sicd.grid.col.uvect_ecf),
            row_ss=sicd.grid.row.ss,
            col_ss=sicd.grid.col.ss,
            first_pixel=ImageCoordinate([sicd.image_data.first_col, sicd.image_data.first_row]),
        )

        projection_set = None
        ugpn = None
        if sicd.grid.type_value == sicd121.ImageGridType.RGAZIM:
            if sicd.image_formation.image_form_algo == sicd121.ImageFormAlgo.PFA:
                ugpn = xyztype_to_ndarray(sicd.pfa.fpn)
                projection_set = PFAProjectionSet(
                    scp_ecf=scp_ecf,
                    polar_ang_poly=poly1d_to_native(sicd.pfa.polar_ang_poly),
                    spatial_freq_sf_poly=poly1d_to_native(sicd.pfa.spatial_freq_sfpoly),
                    coa_time_poly=time_coa_poly,
                    arp_poly=arp_poly,
                )
            elif sicd.image_formation.image_form_algo == sicd121.ImageFormAlgo.RGAZCOMP:
                projection_set = RGAZCOMPProjectionSet(
                    scp_ecf=scp_ecf, az_scale_factor=sicd.rg_az_comp.az_sf, coa_time_poly=time_coa_poly, arp_poly=arp_poly
                )
        elif sicd.grid.type_value == sicd121.ImageGridType.RGZERO:
            projection_set = INCAProjectionSet(
                r_ca_scp=sicd.rma.inca.r_ca_scp,
                inca_time_coa_poly=poly1d_to_native(sicd.rma.inca.time_capoly),
                drate_sf_poly=poly2d_to_native(sicd.rma.inca.drate_sfpoly),
                coa_time_poly=time_coa_poly,
                arp_poly=arp_poly,
            )
        elif sicd.grid.type_value in [
            sicd121.ImageGridType.PLANE,
            sicd121.ImageGridType.XCTYAT,
            sicd121.ImageGridType.XRGYCR,
        ]:
            projection_set = PlaneProjectionSet(
                scp_ecf=scp_ecf,
                image_plane_urow=xyztype_to_ndarray(sicd.grid.row.uvect_ecf),
                image_plane_ucol=xyztype_to_ndarray(sicd.grid.col.uvect_ecf),
                coa_time_poly=time_coa_poly,
                arp_poly=arp_poly,
            )
        else:
            logger.warning(f"SICD image with unknown grid type {sicd.grid.type_value}. No sensor model created.")
            return None

        sicd_sensor_model = SICDSensorModel(
            coord_converter=coord_converter,
            coa_projection_set=projection_set,
            scp_arp=xyztype_to_ndarray(sicd.scpcoa.arppos),
            scp_varp=xyztype_to_ndarray(sicd.scpcoa.arpvel),
            side_of_track=str(sicd.scpcoa.side_of_track.value),
            u_gpn=ugpn,
        )

        return sicd_sensor_model



import logging
import re
from typing import Dict, List, Optional, Tuple

from defusedxml import ElementTree
from osgeo import gdal, gdalconst

from aws.osml.photogrammetry import SensorModel

from .dynamic_range_adjustment import DRAParameters
from .sensor_model_factory import SensorModelFactory, SensorModelTypes
from .typing import RangeAdjustmentType

logger = logging.getLogger(__name__)


def load_gdal_dataset(image_path: str) -> Tuple[gdal.Dataset, Optional[SensorModel]]:
    """
    This function loads a GDAL raster dataset from the path provided and constructs a camera model
    abstraction used to georeference locations on this image.

    :param image_path: the path to the raster data, may be a local path or a virtual file system
        (e.g. /vsis3/...)
    :return: the raster dataset and sensor model
    """
    try:
        logger.info("GDAL attempted to load image: %s", image_path)
        ds = gdal.Open(image_path)
        if ds is None:
            raise RuntimeError("GDAL Unable to load dataset and UseExceptions is not enabled.")
    except RuntimeError:
        logger.info("Skipping: %s - GDAL Unable to Process", image_path)
        raise ValueError("GDAL Unable to Load: {}".format(image_path))

    # Get a GDAL Geo Transform and any available GCPs
    geo_transform = ds.GetGeoTransform(can_return_null=True)
    ground_control_points = ds.GetGCPs()
    proj_wkt = ds.GetProjection()

    # If this image has NITF TREs defined parse them
    parsed_tres = None
    xml_tres = ds.GetMetadata("xml:TRE")
    if xml_tres is not None and len(xml_tres) > 0:
        parsed_tres = ElementTree.fromstring(xml_tres[0])

    # If this image has NITF DES segments read them
    xml_dess = ds.GetMetadata("xml:DES")

    selected_sensor_model_types = [
        SensorModelTypes.AFFINE,
        SensorModelTypes.PROJECTIVE,
        SensorModelTypes.RPC,
        SensorModelTypes.SICD,
        SensorModelTypes.RSM,
    ]
    # Create the best sensor model available
    sensor_model = SensorModelFactory(
        ds.RasterXSize,
        ds.RasterYSize,
        xml_tres=parsed_tres,
        xml_dess=xml_dess,
        geo_transform=geo_transform,
        proj_wkt=proj_wkt,
        ground_control_points=ground_control_points,
        selected_sensor_model_types=selected_sensor_model_types,
    ).build()

    return ds, sensor_model


def get_minmax_for_type(gdal_type: int) -> Tuple[float, float]:
    """
    This function computes the minimum and maximum values that can be stored in a given GDAL pixel type.

    :param gdal_type: the pixel type
    :return: tuple of min, max values
    """
    min_value = 0
    max_value = 255
    if gdal_type == gdalconst.GDT_Byte:
        min_value = 0
        max_value = 2**8 - 1
    elif gdal_type == gdalconst.GDT_UInt16:
        min_value = 0
        max_value = 2**16 - 1
    elif gdal_type == gdalconst.GDT_Int16:
        min_value = -(2**15)
        max_value = 2**15 - 1
    elif gdal_type == gdalconst.GDT_UInt32:
        min_value = 0
        max_value = 2**32 - 1
    elif gdal_type == gdalconst.GDT_Int32:
        min_value = -(2**31)
        max_value = 2**31 - 1
    elif gdal_type == gdalconst.GDT_UInt64:
        min_value = 0
        max_value = 2**64 - 1
    elif gdal_type == gdalconst.GDT_Int64:
        min_value = -(2**63)
        max_value = 2**63 - 1
    elif gdal_type == gdalconst.GDT_Float32:
        min_value = -3.4028235e38
        max_value = 3.4028235e38
    elif gdal_type == gdalconst.GDT_Float64:
        min_value = -1.7976931348623157e308
        max_value = 1.7976931348623157e308
    else:
        logger.warning("Image uses unsupported GDAL datatype {}. Defaulting to [0,255] range".format(gdal_type))

    return min_value, max_value


def get_type_and_scales(
    raster_dataset: gdal.Dataset,
    desired_output_type: Optional[int] = None,
    range_adjustment: RangeAdjustmentType = RangeAdjustmentType.NONE,
) -> Tuple[int, List[List[int]]]:
    """
    Get type and scales of a provided raster dataset

    :param raster_dataset: the raster dataset containing the region
    :param desired_output_type: type to be output after dynamic range adjustments
    :param range_adjustment: the type of pixel scaling effort to

    :return: a tuple containing type and scales
    """
    scale_params = []
    output_type = gdalconst.GDT_Byte
    num_bands = raster_dataset.RasterCount
    for band_num in range(1, num_bands + 1):
        band = raster_dataset.GetRasterBand(band_num)
        band_type = band.DataType
        min_value, max_value = get_minmax_for_type(band_type)

        if desired_output_type is None:
            output_type = band_type
            output_min = min_value
            output_max = max_value
        else:
            output_type = desired_output_type
            output_min, output_max = get_minmax_for_type(desired_output_type)

        # If a range adjustment is requested compute the range of source pixel values that will be mapped to the full
        # output range.
        selected_min = min_value
        selected_max = max_value
        if range_adjustment is not RangeAdjustmentType.NONE:
            # GetStatistics(1,1) means it is ok to approximate but force computation is stats not already available
            stats = band.GetStatistics(1, 1)
            min_value = stats[0]
            max_value = stats[1]

            num_buckets = int(max_value - min_value)
            if band_type == gdalconst.GDT_Float32 or band_type == gdalconst.GDT_Float64:
                num_buckets = 255

            dra = DRAParameters.from_counts(
                counts=band.GetHistogram(
                    buckets=num_buckets, max=max_value, min=min_value, include_out_of_range=1, approx_ok=1
                ),
                first_bucket_value=min_value,
                last_bucket_value=max_value,
            )

            if range_adjustment == RangeAdjustmentType.DRA:
                selected_min = dra.suggested_min_value
                selected_max = dra.suggested_max_value
            elif range_adjustment == RangeAdjustmentType.MINMAX:
                selected_min = dra.actual_min_value
                selected_max = dra.actual_max_value
            else:
                logger.warning(f"Unknown range adjustment selected {range_adjustment}. Skipping.")

        band_scale_parameters = [selected_min, selected_max, output_min, output_max]
        scale_params.append(band_scale_parameters)

    return output_type, scale_params


def get_image_extension(image_path: str) -> str:
    """
    Get the image extension based on the provided image path

    :param image_path: an image path

    :return: image extension
    """
    possible_extensions = get_extensions_from_driver(image_path)
    selected_extension = select_extension(image_path, possible_extensions)
    image_extension = normalize_extension(selected_extension)
    logger.info("Image extension: {}".format(image_extension))
    return image_extension


def select_extension(image_path: str, possible_extensions: List[str]) -> str:
    """
    Check to see if provided image path contains a known possible extensions

    :param image_path: an image path
    :param possible_extensions: list of possible extensions

    :return: select extension
    """
    selected_extension = "UNKNOWN"
    for i, possible_extension in enumerate(possible_extensions):
        if i == 0:
            selected_extension = possible_extension.upper()
        elif f".{possible_extension}".upper() in image_path.upper():
            selected_extension = possible_extension.upper()
    return selected_extension


def normalize_extension(unnormalized_extension: str) -> str:
    """
    Convert the extension into a proper formatted string

    :param unnormalized_extension: an unnormalized extension

    :return: normalized extension
    """
    normalized_extension = unnormalized_extension.upper()
    if re.search(r"ni?tf", normalized_extension, re.IGNORECASE):
        normalized_extension = "NITF"
    elif re.search(r"tif{1,2}", normalized_extension, re.IGNORECASE):
        normalized_extension = "TIFF"
    elif re.search(r"jpe?g", normalized_extension, re.IGNORECASE):
        normalized_extension = "JPEG"
    return normalized_extension


def get_extensions_from_driver(image_path: str) -> List[str]:
    """
    Returns a list of driver extensions

    :param image_path: an image path

    :return: driver extensions
    """
    driver_extension_lookup = get_gdal_driver_extensions()
    info = gdal.Info(image_path, format="json")
    driver_long_name = info.get("driverLongName")
    return driver_extension_lookup.get(driver_long_name, [])


def get_gdal_driver_extensions() -> Dict[str, List]:
    """
    Returns a list of gdal driver extensions

    :return: gdal driver extensions
    """
    driver_lookup = {}
    for i in range(gdal.GetDriverCount()):
        drv = gdal.GetDriver(i)
        driver_name = drv.GetMetadataItem(gdal.DMD_LONGNAME)
        driver_extensions = drv.GetMetadataItem(gdal.DMD_EXTENSIONS)
        if driver_extensions:
            extension_list = driver_extensions.strip().split(" ")
            driver_lookup[driver_name] = extension_list
    return driver_lookup



from typing import List, Optional


class DRAParameters:
    """
    This class manages a set of parameters used to perform a Dynamic Range Adjustment that is applied when
    converting imagery pixel values (e.g. 11-bit per pixel panchromatic imagery to an 8-bit per pixel grayscale).
    """

    def __init__(
        self, suggested_min_value: float, suggested_max_value: float, actual_min_value: float, actual_max_value: float
    ):
        """
        Constructor for this class.

        :param suggested_min_value: suggested minimum value of the relevant pixel range
        :param suggested_max_value: suggested maximum value of the relevant pixel range
        :param actual_min_value: actual minimum value of pixels in the image
        :param actual_max_value: actual maximum value of pixels in the image
        """
        self.suggested_min_value = suggested_min_value
        self.suggested_max_value = suggested_max_value
        self.actual_min_value = actual_min_value
        self.actual_max_value = actual_max_value

    @staticmethod
    def from_counts(
        counts: List[float],
        first_bucket_value: Optional[float] = None,
        last_bucket_value: Optional[float] = None,
        min_percentage: float = 0.02,
        max_percentage: float = 0.98,
        a: float = 0.2,
        b: float = 0.4,
    ) -> "DRAParameters":
        """
        This static factory method computes a new set of DRA parameters given a histogram of pixel values.

        :param counts: histogram of the pixel values
        :param first_bucket_value: pixel value of the first bucket, defaults to 0
        :param last_bucket_value: pixel value of the last bucket, defaults to bucket index
        :param min_percentage: set point for low intensity pixels that may be outliers
        :param max_percentage: set point for high intensity pixels that may be outliers
        :param a: weighting factor for the low intensity range
        :param b: weighting factor for the high intensity range
        :return: a set of DRA parameters containing recommended and actual ranges of values
        """
        num_histogram_bins = len(counts)
        if not first_bucket_value:
            first_bucket_value = 0
        if not last_bucket_value:
            last_bucket_value = num_histogram_bins

        # Find the first and last non-zero counts
        actual_min_value = 0
        while actual_min_value < num_histogram_bins and counts[actual_min_value] == 0:
            actual_min_value += 1

        actual_max_value = num_histogram_bins - 1
        while actual_max_value > 0 and counts[actual_max_value] == 0:
            actual_max_value -= 1

        # Compute the cumulative distribution
        cumulative_counts = counts.copy()
        for i in range(1, len(cumulative_counts)):
            cumulative_counts[i] = cumulative_counts[i] + cumulative_counts[i - 1]

        # Find the values that exclude the lowest and highest percentages of the counts.
        # This identifies the range that contains most of the pixels while excluding outliers.
        max_counts = cumulative_counts[-1]
        low_threshold = min_percentage * max_counts
        e_min = 0
        while cumulative_counts[e_min] < low_threshold:
            e_min += 1

        high_threshold = max_percentage * max_counts
        e_max = num_histogram_bins - 1
        while cumulative_counts[e_max] > high_threshold:
            e_max -= 1

        min_value = max([actual_min_value, e_min - a * (e_max - e_min)])
        max_value = min([actual_max_value, e_max + b * (e_max - e_min)])

        value_step = (last_bucket_value - first_bucket_value) / num_histogram_bins
        return DRAParameters(
            suggested_min_value=min_value * value_step + first_bucket_value,
            suggested_max_value=max_value * value_step + first_bucket_value,
            actual_min_value=actual_min_value * value_step + first_bucket_value,
            actual_max_value=actual_max_value * value_step + first_bucket_value,
        )

    def __repr__(self):
        return (
            f"DRAParameters(min_value={self.suggested_min_value}, "
            f"max_value={self.suggested_max_value}, "
            f"e_first={self.actual_min_value}, "
            f"e_last={self.actual_max_value}, "
            f")"
        )



from typing import Any, Dict, List, Optional

import numpy as np

from .coordinates import GeodeticWorldCoordinate, ImageCoordinate
from .elevation_model import ElevationModel
from .sensor_model import SensorModel
from .transforms import ProjectiveTransform


class ProjectiveSensorModel(SensorModel):
    """
    This sensor model is used when we have a set of 2D tie point correspondences (longitude, latitude) -> (x, y) for
    an image.
    """

    def __init__(
        self,
        world_coordinates: List[GeodeticWorldCoordinate],
        image_coordinates: List[ImageCoordinate],
    ) -> None:
        """
        This constructor estimates a projective transform given the image correspondences.

        :param world_coordinates: a list of world points
        :param image_coordinates: the corresponding list of locations in the image

        :return: None
        """
        super().__init__()
        src_coordinates = [
            np.array([world_coordinate.longitude, world_coordinate.latitude]) for world_coordinate in world_coordinates
        ]
        dst_coordinates = [image_coordinate.coordinate for image_coordinate in image_coordinates]
        self.lonlat_to_xy_transform = ProjectiveTransform.estimate(np.vstack(src_coordinates), np.vstack(dst_coordinates))

    def image_to_world(
        self,
        image_coordinate: ImageCoordinate,
        elevation_model: Optional[ElevationModel] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> GeodeticWorldCoordinate:
        """
        This function returns the longitude, latitude, elevation world coordinate associated with the x, y coordinate
        of any pixel in the image.

        :param image_coordinate: the x, y image coordinate
        :param elevation_model: optional elevation model used to transform the coordinate
        :param options: optional dictionary of hints, this camera does not support any hints

        :return: the longitude, latitude, elevation world coordinate
        """
        world_coords = self.lonlat_to_xy_transform.inverse(np.array([image_coordinate.coordinate]))
        world_coordinate = GeodeticWorldCoordinate(np.append(world_coords[0], [0]))
        if elevation_model:
            elevation_model.set_elevation(world_coordinate)
        return world_coordinate

    def world_to_image(self, world_coordinate: GeodeticWorldCoordinate) -> ImageCoordinate:
        """
        This function returns the x, y image coordinate associated with a given longitude, latitude, elevation world
        coordinate.

        :param world_coordinate: the longitude, latitude, elevation world coordinate

        :return: the x, y image coordinate
        """
        image_coords = self.lonlat_to_xy_transform.forward(np.array([world_coordinate.coordinate[0:2]]))
        image_coordinate = ImageCoordinate(image_coords[0])
        return image_coordinate



# Telling flake8 to not flag errors in this file. It is normal that these classes are imported but not used in an
# __init__.py file.
# flake8: noqa
"""
The gdal package contains utilities that assist with loading imagery and metadata using the OSGeo GDAL library.
"""

from .gdal_config import GDALConfigEnv, set_gdal_default_configuration
from .gdal_dem_tile_factory import GDALDigitalElevationModelTileFactory
from .gdal_utils import get_image_extension, get_type_and_scales, load_gdal_dataset
from .nitf_des_accessor import NITFDESAccessor
from .sensor_model_factory import ChippedImageInfoFacade, SensorModelFactory, SensorModelTypes
from .typing import GDALCompressionOptions, GDALImageFormats, RangeAdjustmentType

__all__ = [
    "set_gdal_default_configuration",
    "load_gdal_dataset",
    "get_image_extension",
    "get_type_and_scales",
    "GDALCompressionOptions",
    "GDALConfigEnv",
    "GDALDigitalElevationModelTileFactory",
    "GDALImageFormats",
    "RangeAdjustmentType",
    "NITFDESAccessor",
    "ChippedImageInfoFacade",
    "SensorModelFactory",
    "SensorModelTypes",
]



from typing import Any, Dict, List, Optional

import numpy as np

from .coordinates import GeodeticWorldCoordinate, ImageCoordinate
from .elevation_model import ElevationModel
from .sensor_model import SensorModel
from .transforms import ProjectiveTransform


class ChippedImageSensorModel(SensorModel):
    """
    This sensor model should be used when we have pixels for only a portion of the image, but we have a sensor model
    that describes the full image. In this case the image coordinates need to be converted to those of the full
    image before being used by the full sensor model.
    """

    def __init__(
        self,
        original_image_coordinates: List[ImageCoordinate],
        chipped_image_coordinates: List[ImageCoordinate],
        full_image_sensor_model: SensorModel,
    ) -> None:
        """
        Construct a chipped image sensor model given the coordinates of the chip in the original image and the
        coordinates of the chip itself. Normally the chip coordinates will be based on the size of the chipped image
        (i.e. [0,0], [0, height] ...). It is important to make sure the coordinates are in the same order.

        Note that this formulation also allows the chipped images to be scaled differently than the original.

        :param original_image_coordinates: locations in image related to chipped image bounds
        :param chipped_image_coordinates: bounds of the chipped image
        :param full_image_sensor_model: the sensor model for the full image

        :return: None
        """
        super().__init__()
        self.full_image_sensor_model = full_image_sensor_model
        src_coordinates = [image_coordinate.coordinate for image_coordinate in original_image_coordinates]
        dst_coordinates = [image_coordinate.coordinate for image_coordinate in chipped_image_coordinates]
        self.full_to_chip_transform = ProjectiveTransform.estimate(np.vstack(src_coordinates), np.vstack(dst_coordinates))

    def image_to_world(
        self,
        image_coordinate: ImageCoordinate,
        elevation_model: Optional[ElevationModel] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> GeodeticWorldCoordinate:
        """
        This function returns the longitude, latitude, elevation world coordinate associated with the x, y coordinate
        of any pixel in the image.

        :param image_coordinate: the x, y image coordinate
        :param elevation_model: an elevation model used to transform the coordinate
        :param options: a dictionary of options that will be passed on to the full image sensor model

        :return: the longitude, latitude, elevation world coordinate
        """
        full_coords = self.full_to_chip_transform.inverse(np.array([image_coordinate.coordinate]))
        full_image_coordinate = ImageCoordinate(full_coords[0])
        return self.full_image_sensor_model.image_to_world(
            full_image_coordinate, elevation_model=elevation_model, options=options
        )

    def world_to_image(self, world_coordinate: GeodeticWorldCoordinate) -> ImageCoordinate:
        """
        This function returns the x, y image coordinate associated with a given longitude, latitude, elevation world
        coordinate.

        :param world_coordinate: the longitude, latitude, elevation world coordinate

        :return: the x, y image coordinate
        """
        full_image_coordinate = self.full_image_sensor_model.world_to_image(world_coordinate)
        chip_coords = self.full_to_chip_transform.forward(np.array([full_image_coordinate.coordinate]))
        chipped_image_coordinate = ImageCoordinate(chip_coords[0])
        return chipped_image_coordinate



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from .coordinates import GeodeticWorldCoordinate


@dataclass
class ElevationRegionSummary:
    """
    This class contains a general summary of an elevation tile.
    """

    min_elevation: float
    max_elevation: float
    no_data_value: int
    post_spacing: float


class ElevationModel(ABC):
    """
    An elevation model associates a height z for a given x, y of a world coordinate. It typically provides information
    about the terrain associated with longitude, latitude locations of an ellipsoid, but it can also be used to model
    surfaces for other ground domains.
    """

    def __init__(self) -> None:
        pass

    @abstractmethod
    def set_elevation(self, world_coordinate: GeodeticWorldCoordinate) -> None:
        """
        This method updates the elevation component of a world coordinate to match the surface elevation at
        longitude, latitude.

        :param world_coordinate: the coordinate to update

        :return: None
        """

    @abstractmethod
    def describe_region(self, world_coordinate: GeodeticWorldCoordinate) -> Optional[ElevationRegionSummary]:
        """
        Get a summary of the region near the provided world coordinate

        :param world_coordinate: the coordinate at the center of the region of interest
        :return: the summary information
        """


class ConstantElevationModel(ElevationModel):
    """
    A constant elevation model with a single value for all longitude, latitude.
    """

    def __init__(self, constant_elevation: float) -> None:
        """
        Constructs the constant elevation model.

        :param constant_elevation: the elevation value for all longitude, latitude

        :return: None
        """
        super().__init__()
        self.constant_elevation = constant_elevation

    def set_elevation(self, world_coordinate: GeodeticWorldCoordinate) -> None:
        """
        Updates world coordinate's elevation to match the constant elevation.

        :param world_coordinate: the coordinate to update

        :return: None
        """
        world_coordinate.elevation = self.constant_elevation

    def describe_region(self, world_coordinate: GeodeticWorldCoordinate) -> Optional[ElevationRegionSummary]:
        """
        Get a summary of the region near the provided world coordinate

        :param world_coordinate: the coordinate at the center of the region of interest
        :return: [min elevation value, max elevation value, no elevation data value, post spacing]
        """
        return ElevationRegionSummary(
            min_elevation=self.constant_elevation,
            max_elevation=self.constant_elevation,
            no_data_value=-32767,
            post_spacing=30.0,
        )



from abc import ABC
from enum import Enum
from math import floor, pi, radians, sqrt
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from scipy.optimize import minimize

from .coordinates import (
    GeodeticWorldCoordinate,
    ImageCoordinate,
    WorldCoordinate,
    geocentric_to_geodetic,
    geodetic_to_geocentric,
)
from .elevation_model import ConstantElevationModel, ElevationModel
from .math_utils import equilateral_triangle
from .sensor_model import SensorModel, SensorModelOptions

# TODO: Add Support for Grid Based RSM Sensor Models
# TODO: Add Support for Adjustable RSM Sensor Models
# TODO: Add Support for Error Assessments
# TODO: Add typing for ArrayLike inputs once Numpy upgraded to 1.20+
# from numpy.typing import ArrayLike


class RSMImageDomain:
    """
    This RSM image domain is a rectangle defined by the minimum and maximum discrete row coordinate values, and the
    minimum and maximum discrete column coordinate values. It is typically constructed from values in the NITF RSMID
    TRE. For more information see section 5.5 of STDI-0002 Volume 1 Appendix U.
    """

    def __init__(
        self,
        min_row: int,
        max_row: int,
        min_column: int,
        max_column: int,
    ) -> None:
        """
        Construct an RSM Image domain from parameters

        :param min_row: min row
        :param max_row: max row
        :param min_column: min column
        :param max_column: max column

        :return: None
        """
        self.min_row = min_row
        self.max_row = max_row
        self.min_column = min_column
        self.max_column = max_column


class RSMGroundDomainForm(Enum):
    """
    The RSMGroundDomainForm defines how world coordinates (x, y, z) should be interpreted in this sensor model.

    If geodetic, X, Y, and Z, correspond to longitude, latitude, and height above the ellipsoid, respectively.
    Longitude is specified east of the prime meridian, and latitude is specified north of the equator. Units for X, Y,
    and Z, are radians, radians, and meters, respectively. The range for Y is (-pi/2 to pi/2). The range for X is
    (-pi to pi) when GEODETIC, and (0 to 2pi) when GEODETIC_2PI. The latter is specified when the RSM ground domain
    contains a longitude value near pi radians.

    If RECTANGULAR, X, Y, and Z correspond to a coordinate system that is defined as an offset from and rotation
    about the WGS 84 Rectangular coordinate system.

    For more information see the GRNDD TRE field definition in section 5.9 of STDI-0002 Volume 1 Appendix U.
    """

    GEODETIC = "G"
    GEODETIC_2PI = "H"
    RECTANGULAR = "R"


class RSMGroundDomain:
    """
    The RSM ground domain is an approximation of the ground area where the RSM representation is valid. It is a solid
    in three-dimensional space bounded by a hexahedron with quadrilateral faces specified using eight
    three-dimensional vertices.

    It is typically constructed from values in the NITF RSMID TRE. For more information see section 5.6 of
    STDI-0002 Volume 1 Appendix U.
    """

    def __init__(
        self,
        ground_domain_form: RSMGroundDomainForm,
        ground_domain_vertices: List[WorldCoordinate],
        rectangular_coordinate_origin: Optional[WorldCoordinate] = None,
        rectangular_coordinate_unit_vectors: Optional[Any] = None,
        ground_reference_point: Optional[WorldCoordinate] = None,
    ) -> None:
        """
        Construct a RSM ground domain from parameters typically found in the RSMID TRE.

        TODO: Change the type hint for rectangular_coordinate_unit_vectors to Optional[ArrayLike] once Numpy > 1.20

        :param ground_domain_form: form of coordinates used by this sensor model
        :param ground_domain_vertices: collection of vertices defining the valid model area
        :param rectangular_coordinate_origin: optional origin of the rectangular system
        :param rectangular_coordinate_unit_vectors: optional unit vectors to define rectangular system
        :param ground_reference_point: optional reference point for this image

        :return: None
        """
        if len(ground_domain_vertices) != 8:
            raise ValueError(
                f"Incorrect number of vertices in RSMGroundDomain. " f"{len(ground_domain_vertices)} provided, 8 expected."
            )

        if ground_domain_form == RSMGroundDomainForm.RECTANGULAR and (
            rectangular_coordinate_origin is None or rectangular_coordinate_unit_vectors is None
        ):
            raise ValueError("Rectangular coordinate system for RSMGroundDomain has not been defined")

        self.ground_domain_form = ground_domain_form
        self.ground_domain_vertices = ground_domain_vertices
        self.rectangular_coordinate_origin = rectangular_coordinate_origin
        if rectangular_coordinate_unit_vectors is not None:
            self.rectangular_coordinate_unit_vectors = np.array(rectangular_coordinate_unit_vectors, dtype=np.float64)
            self.rectangular_coordinate_unit_vectors_inverse = np.linalg.inv(self.rectangular_coordinate_unit_vectors)
        self.ground_reference_point = ground_reference_point

        self.geodetic_ground_domain_vertices = [
            self.ground_domain_coordinate_to_geodetic(vertex) for vertex in self.ground_domain_vertices
        ]

        # Calculate a bounding box and average elevation for this ground domain in geodetic coordinates
        elevation_sum = 0.0
        self.geodetic_lonlat_bbox = [float("inf"), float("inf"), float("-inf"), float("-inf")]
        for geodetic_vertex in self.geodetic_ground_domain_vertices:
            elevation_sum += geodetic_vertex.elevation
            self.geodetic_lonlat_bbox[0] = min(self.geodetic_lonlat_bbox[0], geodetic_vertex.longitude)
            self.geodetic_lonlat_bbox[1] = min(self.geodetic_lonlat_bbox[1], geodetic_vertex.latitude)
            self.geodetic_lonlat_bbox[2] = max(self.geodetic_lonlat_bbox[2], geodetic_vertex.longitude)
            self.geodetic_lonlat_bbox[3] = max(self.geodetic_lonlat_bbox[3], geodetic_vertex.latitude)

        # If a ground reference point is included we will set the reference geodetic height to match that point.
        # If that point is not included then we will compute it from the average height of the ground domain
        # hexahedron in reference to the WGS84 ellipsoid. See section 5.7 of STDI-0002 Volume 1 Appendix U.
        if self.ground_reference_point is not None:
            geodetic_ground_reference_point = self.ground_domain_coordinate_to_geodetic(self.ground_reference_point)
            self.default_elevation_model = ConstantElevationModel(geodetic_ground_reference_point.elevation)
        else:
            average_elevation = elevation_sum / len(self.geodetic_ground_domain_vertices)
            self.default_elevation_model = ConstantElevationModel(average_elevation)

    def geodetic_to_ground_domain_coordinate(self, geodetic_coordinate: GeodeticWorldCoordinate) -> WorldCoordinate:
        """
        This function converts WGS-84 geodetic world coordinate into a world coordinate that uses the domain
        coordinate system for this sensor model.

        :param geodetic_coordinate: the WGS-84 longitude, latitude, elevation

        :return: the x, y, z domain coordinate
        """
        if self.ground_domain_form == RSMGroundDomainForm.RECTANGULAR:
            # This shouldn't happen due to the check in the constructor but the type checker is flagging this, so
            # we check again to ensure we have all the values necessary
            if self.rectangular_coordinate_origin is None or self.rectangular_coordinate_unit_vectors is None:
                raise TypeError("Rectangular ground domain missing origin or unit vectors")

            # The ground domain uses a rectangular coordinate system, so we need to convert the input geodetic
            # coordinate (longitude, latitude, elevation) to the rectangular coordinate system using the provided
            # parameters. See STDI-002 Volume 1 Appendix U Section 5.3 for more details.

            # The longitude, latitude, elevation in (radians, radians, meters) needs to be converted to a
            # WGS84 earth centric coordinate system before subtracting off the coordinate system origin. Then
            # the matrix of unit vectors is applied as shown in the specification.
            ecef_world_coordinate = geodetic_to_geocentric(geodetic_coordinate)
            coordinate_relative_to_origin = ecef_world_coordinate.coordinate - self.rectangular_coordinate_origin.coordinate
            ground_domain_coordinate = WorldCoordinate(
                np.dot(
                    self.rectangular_coordinate_unit_vectors,
                    coordinate_relative_to_origin,
                )
            )
        elif self.ground_domain_form == RSMGroundDomainForm.GEODETIC_2PI:
            # The ground domain uses geodetic coordinates but the range of longitude values is [0,2*PI). Adjust the
            # input value to ensure that range assumption is met.
            ground_domain_coordinate = GeodeticWorldCoordinate(
                [
                    (geodetic_coordinate.x + 2 * pi) % (2 * pi),
                    geodetic_coordinate.y,
                    geodetic_coordinate.z,
                ]
            )
        else:
            # The ground domain is using normal geodetic coordinates so there is nothing to do.
            ground_domain_coordinate = geodetic_coordinate

        return ground_domain_coordinate

    def ground_domain_coordinate_to_geodetic(self, ground_domain_coordinate: WorldCoordinate) -> GeodeticWorldCoordinate:
        """
        This function converts an x, y, z coordinate defined in the ground domain of this sensor model into a WGS-84
        longitude, latitude, elevation coordinate.

        :param ground_domain_coordinate: the x, y, z domain coordinate

        :return: the WGS-84 longitude, latitude, elevation coordinate
        """
        if self.ground_domain_form == RSMGroundDomainForm.RECTANGULAR:
            # This shouldn't happen due to the check in the constructor but the type checker is flagging this, so
            # we check again to ensure we have all the values necessary
            if self.rectangular_coordinate_origin is None or self.rectangular_coordinate_unit_vectors is None:
                raise TypeError("Rectangular ground domain missing origin or unit vectors")

            # The ground domain uses a rectangular coordinate system, so we need to convert the input rectangular
            # coordinate (x,y,z) to the geodetic coordinate system (longitude, latitude, elevation) using the inverse
            # of the provided parameters. See STDI-002 Volume 1 Appendix U Section 5.3 for more details.
            ecef_world_coordinate = GeodeticWorldCoordinate(
                np.dot(
                    self.rectangular_coordinate_unit_vectors_inverse,
                    ground_domain_coordinate.coordinate,
                )
                + self.rectangular_coordinate_origin.coordinate
            )
            world_coordinate = geocentric_to_geodetic(ecef_world_coordinate)
        else:
            # The ground domain is using normal geodetic coordinates so there is nothing to do.
            # TODO: Consider if we should adjust longitude to be in the [-PI,PI) range
            world_coordinate = GeodeticWorldCoordinate(ground_domain_coordinate.coordinate)

        return world_coordinate


class RSMContext:
    """
    The RSM context contains information necessary to apply and interpret results from the sensor models on an image.

    This current implementation only covers the ground domain and image domains necessary for georeferencing but it
    can be expanded as needed to support other RSM functions.

    TODO: Implement the TimeContext which can be used to identify the collection time of any x, y image coordinate
    TODO: Implement the IlluminationContext which can be used to predict shadow information on an image
    TODO: Implement the TrajectoryModel which captures the sensors 3D position in relation to the image
    """

    def __init__(self, ground_domain: RSMGroundDomain, image_domain: RSMImageDomain) -> None:
        """
        Construct the overall RSM context from parameters typically found in the RSMID TRE.

        :param ground_domain: the RSM ground domain
        :param image_domain: the RSM image domain

        :return: None
        """
        self.ground_domain = ground_domain
        self.image_domain = image_domain


class RSMPolynomial:
    """
    This is an implementation of a general polynomial that can be applied to a world coordinate (i.e. an x,y,z vector).
    For additional information see Section 7.2 of STDI-0002 Volume 1 Appendix U or Section 10.3.3.1.1 of the Manual of
    Photogrammetry Sixth Edition.
    """

    def __init__(
        self,
        max_power_x: int,
        max_power_y: int,
        max_power_z: int,
        coefficients: List[float],
    ) -> None:
        """
        Construct the polynomial from coefficients and information about the maximum power.
        Note that len(coefficients) must equal (max_power_x +1) * (max_power_y +1) * (max_power_z +1)

        :param max_power_x: maximum power of x
        :param max_power_y: maximum power of y
        :param max_power_z: maximum power of z
        :param coefficients: coefficients orders in sequence defined in the NITF RSMPCA TRE

        :return: None
        """
        # There must be one coefficient for every combination of x, y, z powers. For example a polynomial with
        # a maximum power of 0 for x, y, and z with only have a single coefficient (the constant C). A polynomial with
        # a maximum power of 1 for x, y, and z will have 8 coefficients for all the combinations of x, y, and z up to
        # a power of 1: [C, X, Y, XY, Z, XZ, YZ, XYZ].
        if len(coefficients) != (max_power_x + 1) * (max_power_y + 1) * (max_power_z + 1):
            raise ValueError("Incorrect number of coefficients in RSMPolynomial")

        self.max_power_x = max_power_x
        self.max_power_y = max_power_y
        self.max_power_z = max_power_z
        self.coefficients = coefficients

    def evaluate(self, normalized_world_coordinate: WorldCoordinate) -> float:
        """
        This function evaluates the polynomial for the given world coordinate by summing up the result of applying
        each coefficient to the world coordinate components. Note that these polynomials are usually defined with
        the assumption that the world coordinate has been normalized.

        :param normalized_world_coordinate: the world coordinate

        :return: the resulting value
        """
        result = 0.0
        a_index = 0
        # Black formatter doesn't play well with the **'s wrapped in brackets
        # fmt: off
        for k in range(self.max_power_z + 1):
            for j in range(self.max_power_y + 1):
                for i in range(self.max_power_x + 1):
                    result += (
                        self.coefficients[a_index]
                        * (normalized_world_coordinate.x ** i)
                        * (normalized_world_coordinate.y ** j)
                        * (normalized_world_coordinate.z ** k)
                    )
                    a_index += 1
        # fmt: on
        return result

    def __call__(self, *args, **kwargs):
        """
        This makes the polynomial object callable such that it can be applied to a world coordinate directly e.g.
        result = polynomial(coordinate).

        :param args: the world coordinate
        :param kwargs: no keyword arguments are accepted but this is required for the Callable interface
        :return: the result of the evaluation
        """
        return self.evaluate(args[0])


class RSMLowOrderPolynomial:
    """
    This is an implementation of a "low order" polynomial used when generating coarse image row and column coordinates
    from a world coordinate. For additional information see Section 6.2 of STDI-0002 Volume 1 Appendix U.
    """

    def __init__(self, coefficients: List[float]) -> None:
        """
        Construct a low order polynomial given a set of coefficients that are ordered to match the specific set of
        component powers [0, X, Y, Z, XX, XY, XZ, YY, YZ, ZZ]

        :param coefficients: the low order polynomial coefficients

        :return: None
        """
        if len(coefficients) != 10:
            raise ValueError("Incorrect number of coefficients when constructing RSMLowOderPolynomial")

        self.coefficients = coefficients

    def evaluate(self, world_coordinate: WorldCoordinate):
        """
        This function evaluates the polynomial for the given world coordinate by summing up the result of applying
        each coefficient to the world coordinate components.

        :param world_coordinate: the world coordinate

        :return: the resulting value
        """
        result = 0.0
        result += self.coefficients[0]  # constant
        result += self.coefficients[1] * world_coordinate.x  # X
        result += self.coefficients[2] * world_coordinate.y  # Y
        result += self.coefficients[3] * world_coordinate.z  # Z
        result += self.coefficients[4] * world_coordinate.x * world_coordinate.x  # XX
        result += self.coefficients[5] * world_coordinate.x * world_coordinate.y  # XY
        result += self.coefficients[6] * world_coordinate.x * world_coordinate.z  # XZ
        result += self.coefficients[7] * world_coordinate.y * world_coordinate.y  # YY
        result += self.coefficients[8] * world_coordinate.y * world_coordinate.z  # YZ
        result += self.coefficients[9] * world_coordinate.z * world_coordinate.z  # ZZ
        return result

    def __call__(self, *args, **kwargs):
        """
        This makes the polynomial object callable such that it can be applied to a world coordinate directly e.g.
        result = polynomial(coordinate).

        :param args: the world coordinate
        :param kwargs: no keyword arguments are accepted but this is required for the Callable interface
        :return: the result of the evaluation
        """
        return self.evaluate(args[0])


class RSMSensorModel(SensorModel, ABC):
    """
    This is an abstract base for all sensor models that use the RSM context information.
    """

    def __init__(self, context: RSMContext) -> None:
        """
        Constructor that accepts the RSM context as an input.

        :param context: contextual information describing the collection environment

        :return: None
        """
        super().__init__()
        self.context = context


class RSMPolynomialSensorModel(RSMSensorModel):
    """
    This is an implementation of a Rational Polynomial Camera as defined in section 10.3.3.1.1 of the Manual of
    Photogrammetry Sixth Edition.
    """

    def __init__(
        self,
        context: RSMContext,
        section_row: int,
        section_col: int,
        row_norm_offset: float,
        column_norm_offset: float,
        x_norm_offset: float,
        y_norm_offset: float,
        z_norm_offset: float,
        row_norm_scale: float,
        column_norm_scale: float,
        x_norm_scale: float,
        y_norm_scale: float,
        z_norm_scale: float,
        row_numerator_poly: RSMPolynomial,
        row_denominator_poly: RSMPolynomial,
        column_numerator_poly: RSMPolynomial,
        column_denominator_poly: RSMPolynomial,
    ) -> None:
        """
        This constructs the sensor model using parameters normally found in the NITF RSMPC TRE.

        :param context: contextual information describing the collection environment
        :param section_row: image row section number that the sensor model applies to
        :param section_col: image col section number that the sensor model applies to
        :param row_norm_offset: offset used to normalize/denormalize image row components
        :param column_norm_offset: offset used to normalize/denormalize image column components
        :param x_norm_offset: offset used to normalize/denormalize world x components
        :param y_norm_offset: offset used to normalize/denormalize world y components
        :param z_norm_offset: offset used to normalize/denormalize world z components
        :param row_norm_scale: scale used to normalize/denormalize image row components
        :param column_norm_scale: scale used to normalize/denormalize image column components
        :param x_norm_scale: scale used to normalize/denormalize world x components
        :param y_norm_scale: scale used to normalize/denormalize world y components
        :param z_norm_scale: scale used to normalize/denormalize world z components
        :param row_numerator_poly: polynomial used as the numerator in row calculations
        :param row_denominator_poly: polynomial used as the denominator in row calculations
        :param column_numerator_poly: polynomial used as the numerator in column calculations
        :param column_denominator_poly: polynomial used as the denominator in column calculations

        :return: None
        """
        super().__init__(context)
        self.section_row = section_row
        self.section_col = section_col
        self.row_norm_offset = row_norm_offset
        self.column_norm_offset = column_norm_offset
        self.x_norm_offset = x_norm_offset
        self.y_norm_offset = y_norm_offset
        self.z_norm_offset = z_norm_offset
        self.row_norm_scale = row_norm_scale
        self.column_norm_scale = column_norm_scale
        self.x_norm_scale = x_norm_scale
        self.y_norm_scale = y_norm_scale
        self.z_norm_scale = z_norm_scale
        self.row_numerator_poly = row_numerator_poly
        self.row_denominator_poly = row_denominator_poly
        self.column_numerator_poly = column_numerator_poly
        self.column_denominator_poly = column_denominator_poly

    def world_to_image(self, geodetic_coordinate: GeodeticWorldCoordinate) -> ImageCoordinate:
        """
        This function transforms a geodetic world coordinate (longitude, latitude, elevation) into an image coordinate
        (x, y).

        :param geodetic_coordinate: the world coordinate (longitude, latitude, elevation)

        :return: the resulting image coordinate (x,y)
        """

        # Convert the geodetic world coordinate into a coordinate in the ground domain then run the ground domain
        # to image function to get the final result.
        world_coordinate = self.context.ground_domain.geodetic_to_ground_domain_coordinate(geodetic_coordinate)
        return self.ground_domain_to_image(world_coordinate)

    def image_to_world(
        self,
        image_coordinate: ImageCoordinate,
        elevation_model: Optional[ElevationModel] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> GeodeticWorldCoordinate:
        """
        This function implements the image to world transform by iteratively invoking world to image within a
        minimization routine to find a matching image coordinate. The longitude and latitude parameters are searched
        independently while the elevation of the world coordinate comes from the surface provided with the ground
        domain.

        :param image_coordinate: the image coordinate (x, y)
        :param elevation_model: an optional elevation model used transform the coordinate
        :param options: optional hints, supports initial_guess and initial_search_distance

        :return: the corresponding world coordinate
        """

        # This is the function we will be minimizing. Given a longitude, latitude coordinate we invoke the
        # world_to_image function to get a projection of that location in the image. Then we compute the
        # distance between that new image location and the input image location. When those locations match then
        # we know we have the world coordinate that corresponds to the input.
        def distance_to_target_coordinate(lonlat_coord: Tuple[float, float]) -> float:
            current_world_coordinate = GeodeticWorldCoordinate([lonlat_coord[0], lonlat_coord[1], 0.0])
            self.context.ground_domain.default_elevation_model.set_elevation(current_world_coordinate)
            if elevation_model:
                elevation_model.set_elevation(current_world_coordinate)
            new_image_coordinate = self.world_to_image(current_world_coordinate)
            return sqrt(
                (image_coordinate.x - new_image_coordinate.x) ** 2 + (image_coordinate.y - new_image_coordinate.y) ** 2
            )

        # Select an initial guess that is at the center of face 1 in the ground domain. Face 1 is defined as the
        # plane V1->V3->V4->V2 so taking a location at the center of the diagonal V1->V4 should start the search off
        # at the center of the ground domain.
        v1 = self.context.ground_domain.geodetic_ground_domain_vertices[0]
        v4 = self.context.ground_domain.geodetic_ground_domain_vertices[3]
        initial_guess = options.get(SensorModelOptions.INITIAL_GUESS) if options is not None else None
        if initial_guess is None:
            initial_guess = np.array([(v1.longitude + v4.longitude) / 2.0, (v1.latitude + v4.latitude) / 2.0])
        if isinstance(initial_guess, List):
            initial_guess = np.array(initial_guess)

        initial_search_distance = options.get(SensorModelOptions.INITIAL_SEARCH_DISTANCE) if options is not None else None
        if initial_search_distance is None:
            initial_search_distance = sqrt(((v1.longitude - v4.longitude) ** 2) + ((v1.latitude - v4.latitude) ** 2))

        # Iteratively adjust the initial guess to minimize the distance to the target image coordinate. We are only
        # allowing the x,y components to vary here and the z is fixed to the elevation model used by the ground
        # domain. The starting simplex is estimated as a triangle centered in the ground domain.
        res = minimize(
            distance_to_target_coordinate,
            initial_guess,
            method="Nelder-Mead",
            bounds=[
                (
                    self.context.ground_domain.geodetic_lonlat_bbox[0],
                    self.context.ground_domain.geodetic_lonlat_bbox[2],
                ),
                (
                    self.context.ground_domain.geodetic_lonlat_bbox[1],
                    self.context.ground_domain.geodetic_lonlat_bbox[3],
                ),
            ],
            options={
                "xatol": radians(0.000001),
                "fatol": 0.5,
                "initial_simplex": equilateral_triangle(initial_guess.tolist(), initial_search_distance),
            },
        )

        # The minimization result is a (longitude,latitude) tuple, so we need to expand it to
        # longitude,latitude,elevation and replace the z component with the height from the elevation model.
        world_coordinate = GeodeticWorldCoordinate(np.append(res.x, 0.0))
        self.context.ground_domain.default_elevation_model.set_elevation(world_coordinate)
        if elevation_model:
            elevation_model.set_elevation(world_coordinate)

        return world_coordinate

    def ground_domain_to_image(self, domain_coordinate: WorldCoordinate) -> ImageCoordinate:
        """
        This function implements the polynomial ground-to-image transform as defined by section 10.3.3.1.1 of the
        Manual of Photogrammetry sixth edition. The world coordinate is first normalized using the offsets and scale
        factors provided. Then the rational polynomial equations are run to produce an x,y image coordinate. Those
        components are then denormalized to find the final image coordinate.

        :param domain_coordinate: the ground domain coordinate (x, y, z)

        :return: the image coordinate (x, y)
        """
        norm_domain_coordinate = self.normalize_world_coordinate(domain_coordinate)
        norm_row = self.row_numerator_poly(norm_domain_coordinate) / self.row_denominator_poly(norm_domain_coordinate)
        norm_column = self.column_numerator_poly(norm_domain_coordinate) / self.column_denominator_poly(
            norm_domain_coordinate
        )
        return self.denormalize_image_coordinate(ImageCoordinate([norm_column, norm_row]))

    def normalize_world_coordinate(self, world_coordinate: WorldCoordinate) -> WorldCoordinate:
        """
        This is a helper function used to normalize a world coordinate for use with the polynomials in this sensor
        model.

        :param world_coordinate: the world coordinate (longitude, latitude, elevation)

        :return: a world coordinate where each component has been normalized
        """
        norm_x = RSMPolynomialSensorModel.normalize(world_coordinate.x, self.x_norm_offset, self.x_norm_scale)
        norm_y = RSMPolynomialSensorModel.normalize(world_coordinate.y, self.y_norm_offset, self.y_norm_scale)
        norm_z = RSMPolynomialSensorModel.normalize(world_coordinate.z, self.z_norm_offset, self.z_norm_scale)
        return WorldCoordinate([norm_x, norm_y, norm_z])

    def denormalize_world_coordinate(self, world_coordinate: WorldCoordinate) -> WorldCoordinate:
        """
        This is a helper function used to denormalize a world coordinate for use with the polynomials in this sensor
        model.

        :param world_coordinate: the normalized world coordinate (longitude, latitude, elevation)

        :return: a world coordinate
        """
        denorm_x = RSMPolynomialSensorModel.denormalize(world_coordinate.x, self.x_norm_offset, self.x_norm_scale)
        denorm_y = RSMPolynomialSensorModel.denormalize(world_coordinate.y, self.y_norm_offset, self.y_norm_scale)
        denorm_z = RSMPolynomialSensorModel.denormalize(world_coordinate.z, self.z_norm_offset, self.z_norm_scale)
        return WorldCoordinate([denorm_x, denorm_y, denorm_z])

    def normalize_image_coordinate(self, image_coordinate: ImageCoordinate) -> ImageCoordinate:
        """
        This is a helper function used to normalize an image coordinate for use in these polynomials.

        :param image_coordinate: the image coordinate (x, y)

        :return: the normalized image coordinate (x, y)
        """
        norm_line = RSMPolynomialSensorModel.normalize(image_coordinate.r, self.row_norm_offset, self.row_norm_scale)
        norm_samp = RSMPolynomialSensorModel.normalize(image_coordinate.c, self.column_norm_offset, self.column_norm_scale)
        return ImageCoordinate([norm_samp, norm_line])

    def denormalize_image_coordinate(self, image_coordinate: ImageCoordinate) -> ImageCoordinate:
        """
        This is a helper function to denormalize an image coordinate after it has been processed by the polynomials.

        :param image_coordinate: the normalized image coordinate (x, y)

        :return: the image coordinate (x, y)
        """
        denorm_line = RSMPolynomialSensorModel.denormalize(image_coordinate.r, self.row_norm_offset, self.row_norm_scale)
        denorm_samp = RSMPolynomialSensorModel.denormalize(
            image_coordinate.c, self.column_norm_offset, self.column_norm_scale
        )
        return ImageCoordinate([denorm_samp, denorm_line])

    @staticmethod
    def normalize(value: float, offset: float, scale: float) -> float:
        """
        This function normalizes a value using an offset and scale using the equations defined in Section 7.2 of
        STDI-0002 Volume 1 Appendix U.

        :param value: the value to be normalized
        :param offset: the normalization offset
        :param scale: the normalization scale

        :return: the normalized value
        """
        return (value - offset) / scale

    @staticmethod
    def denormalize(value: float, offset: float, scale: float) -> float:
        """
        This function denormalizes a value using an offset and scale using the equations defined in Section 7.2 of
        STDI-0002 Volume 1 Appendix U.

        :param value: the normalized value
        :param offset: the normalization offset
        :param scale: the normalization scale

        :return: the denormalized value
        """
        return value * scale + offset


class RSMSectionedPolynomialSensorModel(RSMSensorModel):
    """
    This is an implementation of a sectioned sensor model that splits overall RSM domain into multiple regions each
    serviced by a dedicated sensor model. A low complexity sensor model covering the entire domain is used to first
    identify the general region of the image associated with a world coordinate then the final coordinate transform
    is delegated to a sensor model associated with that section.
    """

    def __init__(
        self,
        context: RSMContext,
        row_num_image_sections: int,
        column_num_image_sections: int,
        row_section_size: float,
        column_section_size: float,
        row_polynomial: RSMLowOrderPolynomial,
        column_polynomial: RSMLowOrderPolynomial,
        section_sensor_models: List[List[SensorModel]],
    ) -> None:
        """
        This constructs the sensor model using parameters normally found in the NITF RSMPI TRE. Note that the
        per-section sensor models should be constructed from their individual RSMPC TREs.

        :param context: contextual information describing the collection environment
        :param row_num_image_sections: the number of row image sections
        :param column_num_image_sections: the number of column image sections
        :param row_section_size: the size of row section
        :param column_section_size: the size of column section
        :param row_polynomial: polynomial used in row calculations
        :param column_polynomial: polynomial used in column calculations
        :param section_sensor_models: the list of sensor models

        :return: None
        """
        super().__init__(context)
        self.row_num_image_sections = row_num_image_sections
        self.column_num_image_sections = column_num_image_sections
        self.row_section_size = row_section_size
        self.column_section_size = column_section_size
        self.row_polynomial = row_polynomial
        self.column_polynomial = column_polynomial
        self.section_sensor_models = section_sensor_models

    def world_to_image(self, geodetic_coordinate: GeodeticWorldCoordinate) -> ImageCoordinate:
        """
        This function transforms a geodetic world coordinate (longitude, latitude, elevation) into an image coordinate
        (x, y).

        :param geodetic_coordinate: the world coordinate (longitude, latitude, elevation)

        :return: the resulting image coordinate (x,y)
        """

        # Use the low order polynomials to approximate an x,y location of this world coordinate in the image
        domain_coordinate = self.context.ground_domain.geodetic_to_ground_domain_coordinate(geodetic_coordinate)
        approximate_row = self.row_polynomial(domain_coordinate)
        approximate_column = self.column_polynomial(domain_coordinate)

        # Find the section of the image containing this approximate coordinate and select the corresponding sensor
        # model
        approximate_image_coordinate = ImageCoordinate([approximate_column, approximate_row])
        column_section_index, row_section_index = self.get_section_index(approximate_image_coordinate)
        section_sensor_model = self.section_sensor_models[row_section_index][column_section_index]

        # Use the selected sensor model to complete the full precision world to image transformation
        return section_sensor_model.world_to_image(geodetic_coordinate)

    def image_to_world(
        self,
        image_coordinate: ImageCoordinate,
        elevation_model: Optional[ElevationModel] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> GeodeticWorldCoordinate:
        """
        This function implements the image to world transform by selecting the sensor model responsible for coordinates
        in the image section and then delegating the image to world calculations to that sensor model.

        :param image_coordinate: the image coordinate (x, y)
        :param elevation_model: optional elevation model used to transform the coordinate
        :param options: optional dictionary of hints passed on to the section sensor models

        :return: the corresponding world coordinate (longitude, latitude, elevation)
        """

        # Find the section of the image containing this coordinate and select the corresponding sensor model
        column_section_index, row_section_index = self.get_section_index(image_coordinate)
        section_camera = self.section_sensor_models[row_section_index][column_section_index]

        # Use the selected sensor model to complete the full precision image to world transformation
        return section_camera.image_to_world(image_coordinate, elevation_model=elevation_model, options=options)

    def get_section_index(self, image_coordinate: ImageCoordinate) -> Tuple[int, int]:
        """
        Use the equations from STDO-0002 Volume 1 Appendix U Section 6.3 to calculate the section of this image
        containing the rough x, y. Note that these equations are slightly different from the documentation since
        those equations produce section numbers that start with 1, and we're starting with 0 to more naturally
        index into the array of sensor models. Note that if the value is outside the normal sections it is clamped
        to use the sensor model from the closest section available.

        :param image_coordinate: the image coordinate (x, y)

        :return: section index (x, y)
        """
        row_section_index = floor((image_coordinate.y - self.context.image_domain.min_row) / self.row_section_size)
        if row_section_index < 0:
            row_section_index = 0
        elif row_section_index >= self.row_num_image_sections:
            row_section_index = self.row_num_image_sections - 1

        column_section_index = floor((image_coordinate.x - self.context.image_domain.min_column) / self.column_section_size)
        if column_section_index < 0:
            column_section_index = 0
        elif column_section_index >= self.column_num_image_sections:
            column_section_index = self.column_num_image_sections - 1

        return column_section_index, row_section_index
